# Copyright (C) 2014-2020 The BET Development Team

from dolfin import*


def solvePoissonRandomField(rand_field, V, f, bcs):
    """
    Solves the poisson equation with a random field :
    (\grad \dot (rand_field \grad(u)) = -f)
    """
    # create the function space
    u = TrialFunction(V)
    v = TestFunction(V)
    L = f * v * dx
    a = inner(rand_field * nabla_grad(u), nabla_grad(v)) * dx
    u = Function(V)
    solve(a == L, u, bcs)
    return u



# Copyright (C) 2014-2020 The BET Development Team


import unittest
import os
import glob
import numpy as np
import numpy.testing as nptest
import bet
import bet.sample as sample
import bet.util as util
import bet.surrogates as surrogates
import bet.sampling.basicSampling as bsam
import bet.calculateP.simpleFunP as simpleFunP
import bet.calculateP.calculateError as calculateError
from bet.Comm import comm, MPI


def linear_model1(parameter_samples):
    Q_map = np.array([[0.506, 0.463], [0.253, 0.918], [0.085, 0.496]])
    QoI_samples = np.dot(parameter_samples, Q_map)
    return QoI_samples


def linear_model2(parameter_samples):
    Q_map = np.array([[0.506], [0.253], [0.085]])
    QoI_samples = np.dot(parameter_samples, Q_map)
    return QoI_samples


def linear_model3(parameter_samples):
    Q_map = np.array([[0.506]])
    QoI_samples = np.dot(parameter_samples, Q_map)
    return QoI_samples


class calculate_error(unittest.TestCase):
    def setUp(self):
        param_ref = np.array([0.5, 0.5, 0.5])
        Q_ref = linear_model2(param_ref)

        sampler = bsam.sampler(linear_model2)
        input_samples = sample.sample_set(3)
        input_samples.set_domain(np.repeat([[0.0, 1.0]], 3, axis=0))
        input_samples = sampler.random_sample_set(rv='uniform', input_obj=input_samples,
                                                  num_samples=1E2)
        disc = sampler.compute_qoi_and_create_discretization(input_samples,
                                                             globalize=True)
        simpleFunP.regular_partition_uniform_distribution_rectangle_scaled(
            data_set=disc, Q_ref=Q_ref, rect_scale=0.5)
        num = disc.check_nums()
        disc._output_sample_set.set_error_estimates(0.01 * np.ones((num, 1)))
        jac = np.zeros((num, 1, 3))
        jac[:, :, :] = np.array([[0.506], [0.253], [0.085]]).transpose()

        disc._input_sample_set.set_jacobians(jac)
        self.disc = disc

    def test_sampling_error(self):
        """
        Testing :meth:`bet.calculateP.calculateError.sampling_error`
        """
        num = self.disc.check_nums()
        neiList = calculateError.cell_connectivity_exact(self.disc)
        for i in range(num):
            self.assertGreater(len(neiList[i]), 0)
        (B_N, C_N) = calculateError.boundary_sets(self.disc, neiList)
        s_error = calculateError.sampling_error(self.disc, exact=True)
        (upper, lower) = s_error.calculate_for_contour_events()
        for x in upper:
            if not np.isnan(x):
                self.assertGreaterEqual(x, 0.0)

        for x in lower:
            if not np.isnan(x):
                self.assertLessEqual(x, 0.0)

        s_set = self.disc._input_sample_set.copy()
        regions_local = np.equal(self.disc._io_ptr_local, 0)
        s_set.set_region_local(regions_local)
        s_set.local_to_global()
        (up, low) = s_error.calculate_for_sample_set_region(s_set,
                                                            1)
        if not np.isnan(up):
            self.assertAlmostEqual(up, upper[0])
        else:
            self.assertTrue(np.isnan(upper[0]))
        if not np.isnan(low):
            self.assertAlmostEqual(low, lower[0])
        else:
            self.assertTrue(np.isnan(lower[0]))

        (up, low) = s_error.calculate_for_sample_set_region(s_set,
                                                            1,
                                                            emulated_set=self.disc._input_sample_set)
        if np.isnan(up):
            self.assertEqual(np.isnan(upper[0]), True)
        else:
            self.assertAlmostEqual(up, upper[0])

        if np.isnan(low):
            self.assertEqual(np.isnan(lower[0]), True)
        else:
            self.assertAlmostEqual(low, lower[0])

        self.disc.set_emulated_input_sample_set(self.disc._input_sample_set)
        (up, low) = s_error.calculate_for_sample_set_region(s_set,
                                                            1)
        if np.isnan(up):
            self.assertEqual(np.isnan(upper[0]), True)
        else:
            self.assertAlmostEqual(up, upper[0])

        if np.isnan(low):
            self.assertEqual(np.isnan(lower[0]), True)
        else:
            self.assertAlmostEqual(low, lower[0])

        # function should set pointer if missing.
        self.disc._emulated_ii_ptr_local = None
        (up, low) = s_error.calculate_for_sample_set_region(s_set,
                                                            1)
        if np.isnan(up):
            self.assertEqual(np.isnan(upper[0]), True)
        else:
            self.assertAlmostEqual(up, upper[0])

        if np.isnan(low):
            self.assertEqual(np.isnan(lower[0]), True)
        else:
            self.assertAlmostEqual(low, lower[0])

    def test_model_error(self):
        """
        Testing :meth:`bet.calculateP.calculateError.model_error`
        """
        num = self.disc.check_nums()
        m_error = calculateError.model_error(self.disc)
        er_est = m_error.calculate_for_contour_events()

        s_set = self.disc._input_sample_set.copy()
        regions_local = np.equal(self.disc._io_ptr_local, 0)
        s_set.set_region_local(regions_local)
        s_set.local_to_global()
        er_est2 = m_error.calculate_for_sample_set_region(s_set,
                                                          1)
        self.assertAlmostEqual(er_est[0], er_est2)
        error_id_sum_local = np.sum(
            self.disc._input_sample_set._error_id_local)
        error_id_sum = comm.allreduce(error_id_sum_local, op=MPI.SUM)
        self.assertAlmostEqual(er_est2, error_id_sum)

        emulated_set = self.disc._input_sample_set
        er_est3 = m_error.calculate_for_sample_set_region(s_set,
                                                          1,
                                                          emulated_set=emulated_set)
        self.assertAlmostEqual(er_est[0], er_est3)
        self.disc.set_emulated_input_sample_set(self.disc._input_sample_set)
        m_error = calculateError.model_error(self.disc)
        er_est4 = m_error.calculate_for_sample_set_region(s_set,
                                                          1)
        self.assertAlmostEqual(er_est[0], er_est4)


class Test_3_to_2(calculate_error):
    """
    Testing :meth:`bet.calculateP.calculateError` on a
    3 to 2 map.
    """

    def setUp(self):
        param_ref = np.array([0.5, 0.5, 0.5])
        Q_ref = linear_model1(param_ref)

        sampler = bsam.sampler(linear_model1)
        input_samples = sample.sample_set(3)
        input_samples.set_domain(np.repeat([[0.0, 1.0]], 3, axis=0))
        input_samples = sampler.random_sample_set(rv='uniform', input_obj=input_samples,
                                                  num_samples=1E3)
        disc = sampler.compute_qoi_and_create_discretization(input_samples,
                                                             globalize=True)
        simpleFunP.regular_partition_uniform_distribution_rectangle_scaled(
            data_set=disc, Q_ref=Q_ref, rect_scale=0.5)
        num = disc.check_nums()
        disc._output_sample_set.set_error_estimates(0.01 * np.ones((num, 2)))
        jac = np.zeros((num, 2, 3))
        jac[:, :, :] = np.array(
            [[0.506, 0.463], [0.253, 0.918], [0.085, 0.496]]).transpose()

        disc._input_sample_set.set_jacobians(jac)
        self.disc = disc


class Test_3_to_1(calculate_error):
    """
    Testing :meth:`bet.calculateP.calculateError` on a
    3 to 1 map.
    """

    def setUp(self):
        param_ref = np.array([0.5, 0.5, 0.5])
        Q_ref = linear_model2(param_ref)

        sampler = bsam.sampler(linear_model2)
        input_samples = sample.sample_set(3)
        input_samples.set_domain(np.repeat([[0.0, 1.0]], 3, axis=0))
        input_samples = sampler.random_sample_set(rv='uniform', input_obj=input_samples,
                                                  num_samples=1E2)
        disc = sampler.compute_qoi_and_create_discretization(input_samples,
                                                             globalize=True)
        simpleFunP.regular_partition_uniform_distribution_rectangle_scaled(
            data_set=disc, Q_ref=Q_ref, rect_scale=0.5)
        num = disc.check_nums()
        disc._output_sample_set.set_error_estimates(0.01 * np.ones((num, 1)))
        jac = np.zeros((num, 1, 3))
        jac[:, :, :] = np.array([[0.506], [0.253], [0.085]]).transpose()

        disc._input_sample_set.set_jacobians(jac)
        self.disc = disc


class Test_1_to_1(calculate_error):
    """
    Testing :meth:`bet.calculateP.calculateError` on a
    1 to 1 map.
    """

    def setUp(self):
        param_ref = np.array([0.5])
        Q_ref = linear_model3(param_ref)

        sampler = bsam.sampler(linear_model3)
        input_samples = sample.sample_set(1)
        input_samples.set_domain(np.repeat([[0.0, 1.0]], 1, axis=0))
        input_samples = sampler.random_sample_set(rv='uniform', input_obj=input_samples,
                                                  num_samples=1E2)
        disc = sampler.compute_qoi_and_create_discretization(input_samples,
                                                             globalize=True)
        simpleFunP.regular_partition_uniform_distribution_rectangle_scaled(
            data_set=disc, Q_ref=Q_ref, rect_scale=0.5)
        num = disc.check_nums()
        disc._output_sample_set.set_error_estimates(0.01 * np.ones((num, 1)))
        jac = np.zeros((num, 1, 1))
        jac[:, :, :] = np.array([[0.506]]).transpose()

        disc._input_sample_set.set_jacobians(jac)
        self.disc = disc



# Copyright (C) 2014-2020 The BET Development Team

r"""
This subpackage provides methods for approximating gradients of
QoI maps and choosing optimal QoIs to use in the inverse problem.

* :mod:`~bet.sensitivity.gradients` provides methods for approximating gradients of QoI maps.
* :mod:`~bet.sensitivity.chooseQoIs` provides methods for choosing optimal QoIs to use in the inverse problem.
"""

__all__ = ['gradients', 'chooseQoIs']



#!/usr/bin/en python

# Copyright (C) 2014-2020 The BET Development Team


from dolfin import *
from numpy import *


class meshDS(object):

    """Docstring for meshDS. """

    def __init__(self, mesh):
        """TODO: to be defined1.

        :mesh: reads a fenics mesh object

        """
        self._mesh = mesh
        self.node_elem = {}  # empty dictionary of node to elements connectivity
        self.edges_elem = {}  # empty dictionary of edges to elements connectivity

        # initialize the mesh and read in the values
        self._mesh.init()
        self._dim = self._mesh.topology().dim()
        self.num_nodes = self._mesh.num_vertices()
        self.num_elements = self._mesh.num_cells()
        self.num_edges = self._mesh.num_edges()

    def getNodes(self):
        """TODO: Docstring for getNodes.
        :returns: num of nodes in the mesh

        """
        return self.num_nodes

    def getElements(self):
        """TODO: Docstring for getElements.
        :returns: number of elements in the mesh

        """
        return self.num_elements

    def getEdges(self):
        """TODO: Docstring for getElements.
        :returns: number of elements in the mesh

        """
        return self.num_edges

    def getElemToNodes(self):
        """TODO: Docstring for getElemToNodes.
        :returns: Elements - Nodes Connectivity array of array

        """
        return self._mesh.cells()

    def getNodesToElem(self):
        """TODO: Docstring for getNodesToElem.
        :returns: returns Nodes to Element connectivity as a dictionary
        where nodes_elem[i] is an array of all the elements attached to node i

        """
        for nodes in entities(self._mesh, 0):
            self.node_elem[nodes.index()] = nodes.entities(self._dim)
        return self.node_elem

    def getElemVCArray(self):
        """TODO: Docstring for getElemVCArray.
        :returns: array of element volume and and an array of element centroid object
        Thus elem_centroid_array[i][0] means the x co-ordinate of the centroid for element number i
        Thus elem_centroid_array[i][1] means the y co-ordinate of the centroid for element number i
        """

        elem_vol_array = empty((self.num_elements), dtype=float)
        elem_centroid_array = empty((self.num_elements), dtype=object)

        cell_indx = 0
        for node_list in self._mesh.cells():
            # First get the cell object corresponding to the cell_indx
            cell_obj = Cell(self._mesh, cell_indx)
            # Find the cell volume and cell centroid
            elem_vol_array[cell_indx] = cell_obj.volume()
            elem_centroid_array[cell_indx] = cell_obj.midpoint()
            # update cell index
            cell_indx = cell_indx + 1
        return elem_vol_array, elem_centroid_array



#! /usr/bin/env python

# Copyright (C) 2014-2020 The BET Development Team

r"""

Creates uniformly distributed samples on a unit cube and evaluates a
3d to 2d nonlinear map. The map has normal error added to it, and the
difference between the true model and the model with error is added
as an error estimate. Jacobians are also added.
Assuming an output probability as uniform on a
rectangle, the stochastic inverse problem is solved. The error in the
solution over a defined box due to sampling is approximated with
error bounds. The error due to solving the model inexactly is also
approximated.

Piecewise constant and piecewise linear (using Jacobians) surrogates
are generated and used to calculate probabilities and error estimates.
Effectivity ratios are calculated.

"""

import numpy as np
import bet.calculateP.simpleFunP as simpleFunP
import bet.calculateP.calculateP as calculateP
import bet.calculateP.calculateError as calculateError
import bet.sample as samp
import bet.sampling.basicSampling as bsam
import bet.surrogates as surrogates
from bet.Comm import comm
from lbModel import lb_model_exact, lb_model


# Define the reference parameter
param_ref = np.array([[0.5, 0.5, 0.5]])
(Q_ref, _) = lb_model_exact(param_ref)

# Interface BET to the approximate model and create discretization object.
sampler = bsam.sampler(lb_model, error_estimates=True, jacobians=True)
input_samples = samp.sample_set(3)
input_samples.set_domain(np.array([[0.0, 1.0],
                                   [0.0, 1.0],
                                   [0.0, 1.0]]))
my_disc = sampler.create_random_discretization(
    "random", input_samples, num_samples=1000)

# Define output probability
rect_domain = np.array([[0.5, 1.5], [1.25, 2.25]])
simpleFunP.regular_partition_uniform_distribution_rectangle_domain(
    data_set=my_disc, rect_domain=rect_domain)

# Make emulated input sets
emulated_inputs = bsam.random_sample_set('uniform',
                                         my_disc._input_sample_set._domain,
                                         num_samples=10001,
                                         globalize=False)

emulated_inputs2 = bsam.random_sample_set('uniform',
                                          my_disc._input_sample_set._domain,
                                          num_samples=10001,
                                          globalize=False)


# Make exact discretization
my_disc_exact = my_disc.copy()
my_disc_exact._output_sample_set._values_local = my_disc._output_sample_set._values_local + \
    my_disc._output_sample_set._error_estimates_local
my_disc_exact._output_sample_set._error_estimates_local = np.zeros(
    my_disc_exact._output_sample_set._error_estimates_local.shape)
my_disc_exact._output_sample_set.local_to_global()
my_disc_exact.set_emulated_input_sample_set(emulated_inputs2)

# Solve stochastic inverse problems
my_disc.set_emulated_input_sample_set(emulated_inputs)
calculateP.prob_with_emulated_volumes(my_disc)
calculateP.prob_with_emulated_volumes(my_disc_exact)

# Set in input space to calculate errors for
s_set = samp.rectangle_sample_set(3)
s_set.setup(maxes=[[0.75, 0.75, 0.75]], mins=[[0.25, 0.25, 0.25]])
s_set.set_region(np.array([0, 1]))

# Calculate true probabilty using linear surrogate on true data
sur = surrogates.piecewise_polynomial_surrogate(my_disc_exact)
sur_disc_lin = sur.generate_for_input_set(emulated_inputs, order=1)
(Pt, _) = sur.calculate_prob_for_sample_set_region(s_set,
                                                   regions=[0])
P_true = Pt[0]
if comm.rank == 0:
    print("True probability: ", P_true)

# Approximate probabilty
P_approx = calculateP.prob_from_sample_set(my_disc._input_sample_set,
                                           s_set)[0]
if comm.rank == 0:
    print("Approximate probability: ", P_approx)

# Calculate sampling error
samp_er = calculateError.sampling_error(my_disc)
(up, low) = samp_er.calculate_for_sample_set_region(s_set,
                                                    region=0, emulated_set=emulated_inputs)
if comm.rank == 0:
    print("Sampling error upper bound: ", up)
    print("Sampling error lower bound: ", low)


# Calculate modeling error
mod_er = calculateError.model_error(my_disc)
er_est = mod_er.calculate_for_sample_set_region(s_set,
                                                region=0, emulated_set=emulated_inputs)
if comm.rank == 0:
    print("Modeling error estimate: ", er_est)

# Make piecewise constant surrogate of discretization
sur = surrogates.piecewise_polynomial_surrogate(my_disc)
sur_disc_constant = sur.generate_for_input_set(emulated_inputs, order=0)
# Calculate probability and error estimate
(P2, er_est2) = sur.calculate_prob_for_sample_set_region(s_set,
                                                         regions=[0])
if comm.rank == 0:
    print("Piecewise constant surrogate probability: ", P2[0])
    print("Piecewise constant error estimate ", er_est2[0])
    print("Piecewise constant corrected probability: ", P2[0] - er_est2[0])
    print("Piecewise constant effectivity ratio: ",
          er_est2[0] / (P2[0] - P_true))

# Make piecewise linear surrogate of discretization
sur_disc_linear = sur.generate_for_input_set(emulated_inputs, order=1)
# Calculate probability and error estimate
(P3, er_est3) = sur.calculate_prob_for_sample_set_region(s_set,
                                                         regions=[0])
if comm.rank == 0:
    print("Piecewise linear surrogate probability: ", P3[0])
    print("Piecewise linear error estimate ", er_est3[0])
    print("Piecewise linear corrected probability: ", P3[0] - er_est3[0])
    print("Piecewise linear effectivity ratio: ",
          er_est3[0] / (P3[0] - P_true))



# Copyright (C) 2014-2020 The BET Development Team
import numpy as np
import bet.sample as samp
import bet.sampling.basicSampling as bsam
import scipy.spatial.distance as ds


class compare:
    """
    This class allows for the statistical distance between probability measures
    to be calculated.
    The probability measures may be defined by Voronoi tesselations, weighted Kernel Density Estimates,
    Gaussian Mixture Models, random variables with known parameters, and multi-dimensional
    normal distributions.
    This object can be thought of as a more flexible version of an abstraction
    of a metric, a measure of distance between two probability measures.
    It ``d(x,y)`` takes two arguments, one to the left (``x``),
    and one to the right (``y``). However, we do not enforce the properties
    that define a formal metric, instead we use the language of statistical distance.
    """
    def __init__(self, set1, set2, inputs=True, set1_init=False, set2_init=False):
        """

        Initialize comparison object.

        :param set1: Object containing left probability measure.
        :type set1: :class:`bet.sample.sample_set` or `bet.sample.discretization`lass:
        :param set2: Object containing left probability measure.
        :type set1: :class:`bet.sample.sample_set` or class:`bet.sample.discretization`
        :param inputs: If set1 and set2 are discretizations, use input sets if True and output if False.
            True by default.
        :type inputs: bool
        :param set1_init: Use initial probability measure for set1 if True. False by default.
        :type set1_init: bool
        :param set2_init: Use initial probability measure for set2 if True. False by default.
        :type set2_init: bool
        """
        self.pdfs1 = None
        self.pdfs2 = None
        self.compare_vals = None
        self.set1_init = set1_init
        self.set2_init = set2_init
        self.pdfs_zero = None

        # Extract sample sets
        if isinstance(set1, samp.discretization):
            if inputs:
                set1 = set1.get_input_sample_set()
            else:
                set1 = set1.get_output_sample_set()

        if isinstance(set2, samp.discretization):
            if inputs:
                set2 = set2.get_input_sample_set()
            else:
                set2 = set2.get_output_sample_set()

        if isinstance(set1, samp.sample_set) and isinstance(set2, samp.sample_set):
            self.set1 = set1
            self.set2 = set2
        else:
            raise samp.wrong_input("Inputs are not of valid form.")

        # Check dimensions
        if self.set1.get_dim() != self.set2.get_dim():
            raise samp.dim_not_matching("The sets do not have the same dimension.")

    def set_compare_set(self, compare_set=10000, compare_factor=0.0):
        """
        Set values where the left and right probability measures should be compared.
        If `compare_set` is of type :class:`bet.sample.sample_set`, then the values from
        that object are used. If `compare_set` is of type :class:`numpy.ndarray`, then the
        values in that array are used. If `compare_set` is of type int, then that number
        of uniformly distributed are sampled from a domain containing all of the values
        for set1 and set2. If compare_factor is set to be greater than 0, then this domain
        is increased by that proportion in every direction. Increasing the size of the
        sampling domain may catch areas of nonzero probability.

        :param compare_set: Set containing values on which to compare.
        :type compare_set: :class:`bet.sample.sample_set`, :class:`numpy.ndarray`, or int 10000 by default.
        :param compare_factor: Proportion to increase domain for sampling. Only used if `compare_set` is type int.
            0 by default.
        :type compare_factor: float
        """
        # Extract values to evaluate the probability measures.
        if isinstance(compare_set, samp.sample_set):
            if compare_set.get_dim() == self.set1.get_dim():
                compare_set.local_to_global()
                self.compare_vals = compare_set.get_values()
            else:
                raise samp.dim_not_matching("The sets do not have the same dimension.")
        elif isinstance(compare_set, np.ndarray):
            if compare_set.shape[1] == self.set1.get_dim():
                self.compare_vals = compare_set
            else:
                raise samp.dim_not_matching("The sets do not have the same dimension.")
        elif isinstance(compare_set, int):
            # Find bounds
            combined = np.vstack((self.set1.get_values(), self.set2.get_values()))
            mins = np.min(combined, axis=0)
            maxes = np.max(combined, axis=0)

            # Perform uniform random sampling.
            rv = []
            for i in range(self.set1.get_dim()):
                rv_loc = ['uniform', {}]
                delt = compare_factor * (maxes[i] - mins[i])
                rv_loc[1]['loc'] = mins[i] - delt
                rv_loc[1]['scale'] = maxes[i] - mins[i] + delt
            unif_set = bsam.random_sample_set(rv=rv_loc, input_obj=self.set1.get_dim(),
                                              num_samples=compare_set)
            self.compare_vals = unif_set.get_values()
        else:
            raise samp.wrong_input("Inputs are not of valid form.")

    def evaluate_pdfs(self):
        """
        Evaluate probability density functions associated with the probability measures at
        the comparison points.
        """
        if self.set1_init:
            self.pdfs1 = self.set1.pdf_init(self.compare_vals)
        else:
            self.pdfs1 = self.set1.pdf(self.compare_vals)

        if self.set2_init:
            self.pdfs2 = self.set2.pdf_init(self.compare_vals)
        else:
            self.pdfs2 = self.set2.pdf(self.compare_vals)

        sup1 = np.equal(self.pdfs1, 0.0)
        sup2 = np.equal(self.pdfs2, 0.0)
        self.pdfs_zero = np.sum(np.logical_and(sup1, sup2))

    def distance(self, functional='tv', normalize=False, **kwargs):
        """
        Compute the statistical distance between the probability measures
        evaluated at the comparison points.

        :param functional: functional defining type of statistical distance
        :type functional: str or a function that takes in two lists/arrays and returns
            a scalar value (measure of similarity). Accepted strings are 'tv' (total variation) the
            default, 'kl' (Kullback-Leibler),
            'mink' (minkowski), '2' (Euclidean norm), and 'hell' (Hellinger distance).
        :param normalize: whether or not to normalize the distance
        :type normalize: bool
        :param kwargs: Keyword arguments for `functional`.

        :rtype: float
        :returns: The statistical distance

        """
        # Check inputs
        if self.compare_vals is None:
            raise samp.wrong_input("Compare set needed.")
        if self.pdfs1 is None or self.pdfs2 is None:
            self.evaluate_pdfs()
        if normalize:
            self.pdfs1 = self.pdfs1 / np.sum(self.pdfs1)
            self.pdfs2 = self.pdfs2 / np.sum(self.pdfs2)
            factor = 1.0
        else:
            factor = 1.0 / (self.pdfs1.shape[0])


        if functional in ['tv', 'totvar',
                          'total variation', 'total-variation', '1']:
            dist = factor * ds.minkowski(self.pdfs1, self.pdfs2, 1, w=0.5, **kwargs)
        elif functional in ['mink', 'minkowski']:
            dist = factor * ds.minkowski(self.pdfs1, self.pdfs2, **kwargs)
        elif functional in ['norm']:
            dist = factor * ds.norm(self.pdfs1 - self.pdfs2, **kwargs)
        elif functional in ['euclidean', '2-norm', '2']:
            dist = (factor ** 0.5) * ds.minkowski(self.pdfs1, self.pdfs2, 2, **kwargs)
        elif functional in ['sqhell', 'sqhellinger']:
            dist = factor * ds.sqeuclidean(np.sqrt(self.pdfs1), np.sqrt(self.pdfs2)) / 2.0
        elif functional in ['hell', 'hellinger']:
            return np.sqrt(self.distance('sqhell'))
        elif functional in ['kl', 'k-l', 'kullback-leibler', 'entropy']:
            from scipy.stats import entropy as kl_div
            dist = kl_div(self.pdfs1, self.pdfs2, **kwargs)
        else:
            dist = functional(self.pdfs1, self.pdfs2, **kwargs)
        return dist

    def distance_marginal(self, i, interval=None, num_points=1000, compare_factor=0.0, normalize=False,
                          functional='tv', **kwargs):
        """
        Compute the statistical distance between the marginals of the probability measures
        evaluated at equally spaced points on an interval. If the interval is not defined,
        one is computed by the maximum and minimum values. This domain is extended by the proportion
        set by `compare_factor`.

        :param i: index of the marginal
        :type i: int
        :param interval: interval over which to integrate. None by default.
        :type interval: list, tuple, or :class:`numpy.ndarray`
        :param num_points: number of evaluation points. 1000 by default.
        :type num_points: int
        :param compare_factor: Proportion to increase domain. Only used if
            `interval` is None. 0 by default.
        :type compare_factor: float
        :param normalize: whether or not to normalize the probabilities to sum to 1
        :type normalize: bool
        :param functional: functional defining type of statistical distance
        :type functional: str or a function that takes in two lists/arrays and returns
            a scalar value (measure of similarity). Accepted strings are 'tv' (total variation), 'kl' (Kullback-Leibler)
            'mink' (minkowski), '2' (Euclidean norm), and 'hell' (Hellinger distance).
        :param kwargs: Keyword arguments for `functional`.

        :rtype: float
        :returns: The statistical distance

        """
        x = None
        if interval is None:
            if self.set1.get_domain() is not None and self.set2.get_domain() is not None:
                min1 = min(self.set1.get_domain()[i, 0], self.set1.get_domain()[i, 0])
                max1 = min(self.set1.get_domain()[i, 1], self.set1.get_domain()[i, 1])
                if min1 != -np.inf and max1 != np.inf:
                    delt = compare_factor * (max1 - min1)
                    x = np.linspace(min1-delt, max1+delt, num_points)
            if x is None:
                combined = np.vstack((self.set1.get_values()[:, i], self.set2.get_values()[:, i]))
                min1 = np.min(combined)
                max1 = np.max(combined)
                delt = compare_factor * (max1 - min1)
                x = np.linspace(min1 - delt, max1 + delt, num_points)
        else:
            x = np.linspace(interval[0], interval[1], num_points)

        if self.set1_init:
            pdfs1 = self.set1.marginal_pdf_init(x, i)
        else:
            pdfs1 = self.set1.marginal_pdf(x, i)

        if self.set2_init:
            pdfs2 = self.set2.marginal_pdf_init(x, i)
        else:
            pdfs2 = self.set2.marginal_pdf(x, i)

        if normalize:
            pdfs1 = pdfs1 / np.sum(pdfs1)
            pdfs2 = pdfs2 / np.sum(pdfs2)
            factor = 1.0
        else:
            factor = 1.0 / (pdfs1.shape[0]) * (x[-1] - x[0])

        if functional in ['tv', 'totvar',
                          'total variation', 'total-variation', '1']:
            dist = factor * ds.minkowski(pdfs1, pdfs2, 1, w=0.5, **kwargs)
        elif functional in ['mink', 'minkowski']:
            dist = ds.minkowski(pdfs1, pdfs2, **kwargs)
        elif functional in ['norm']:
            dist = ds.norm(pdfs1 - pdfs2, **kwargs)
        elif functional in ['euclidean', '2-norm', '2']:
            dist = (factor ** 0.5) * ds.minkowski(pdfs1, pdfs2, 2, **kwargs)
        elif functional in ['sqhell', 'sqhellinger']:
            dist = 0.5 * factor * (ds.minkowski(np.sqrt(pdfs1), np.sqrt(pdfs2), 2, **kwargs) ** 2.0)
        elif functional in ['hell', 'hellinger']:
            dist = (0.5 * factor * (ds.minkowski(np.sqrt(pdfs1), np.sqrt(pdfs2), 2, **kwargs) ** 2.0)) ** 0.5
        elif functional in ['kl', 'k-l', 'kullback-leibler', 'entropy']:
            from scipy.stats import entropy as kl_div
            dist = kl_div(pdfs1, pdfs2, **kwargs)
        else:
            dist = functional(pdfs1, pdfs2, **kwargs)

        return dist

    def distance_marginal_quad(self, i, interval=None, compare_factor=0.0,
                               functional='tv', **kwargs):
        """
        Compute the  statistical distance between the marginals of the probability measures
        by integrating using `scipy.integrate.quadrature`.. If the interval is not defined,
        one is computed by the maximum and minimum values. This domain is extended by the proportion
        set by `compare_factor`.

        :param i: index of the marginal
        :type i: int
        :param interval: interval over which to integrate. None by default.
        :type interval: list, tuple, or :class:`numpy.ndarray`
        :param compare_factor: Proportion to increase domain. Only used if
            `interval` is None. 0 by default.
        :type compare_factor: float
        :param functional: functional defining type of statistical distance
        :type functional: str or a function that takes in two lists/arrays and returns
            a scalar value (measure of similarity). Accepted strings are 'tv' (total variation),
            'mink' (minkowski), '2' (Euclidean norm), 'kl' (Kullback-Leibler) and 'hell' (Hellinger distance).
        :param kwargs: Keyword arguments for `scipy.integrate.quadrature`.

        :rtype: float
        :returns: The statistical distance

        """
        from scipy.integrate import quadrature
        if interval is None:
            if self.set1.get_domain() is not None and self.set2.get_domain() is not None:
                min1 = min(self.set1.get_domain()[i, 0], self.set1.get_domain()[i, 0])
                max1 = min(self.set1.get_domain()[i, 1], self.set1.get_domain()[i, 1])
                if min1 != -np.inf and max1 != np.inf:
                    delt = compare_factor * (max1 - min1)
                    interval = [min1-delt, max1 + delt]
            if interval is None:
                combined = np.vstack((self.set1.get_values()[:, i], self.set2.get_values()[:, i]))
                min1 = np.min(combined)
                max1 = np.max(combined)
                delt = compare_factor * (max1 - min1)
                interval = [min1 - delt, max1 + delt]
                
        if self.set1_init:
            pdf1 = self.set1.marginal_pdf_init
        else:
            pdf1 = self.set1.marginal_pdf

        if self.set2_init:
            pdf2 = self.set2.marginal_pdf_init
        else:
            pdf2 = self.set2.marginal_pdf

        if functional in ['tv', 'totvar',
                          'total variation', 'total-variation', '1']:
            def error(x):
                return np.abs(pdf1(x, i) - pdf2(x, i))
            return 0.5 * quadrature(error, interval[0], interval[1], **kwargs)[0]
        elif functional in ['euclidean', '2-norm', '2']:
            def error(x):
                return (pdf1(x, i) - pdf2(x, i))**2
            return (quadrature(error, interval[0], interval[1], **kwargs)[0])**0.5
        elif functional in ['norm']:
            def error(x):
                return pdf1(x, i) - pdf2(x, i)

            return quadrature(error, interval[0], interval[1], **kwargs)[0]
        elif functional in ['sqhell', 'sqhellinger']:
            def error(x):
                return 0.5 * (np.sqrt(pdf1(x, i)) - np.sqrt(pdf2(x, i)))**2
            return quadrature(error, interval[0], interval[1], **kwargs)[0]
        elif functional in ['hell', 'hellinger']:
            return np.sqrt(self.distance_marginal_quad(i, interval, compare_factor=0,
                                                       functional="sqhell", **kwargs))
        elif functional in ['kl', 'k-l', 'kullback-leibler', 'entropy']:
            def error(x):
                return pdf1(x, i) * np.log(pdf1(x, i)/pdf2(x, i))

            return quadrature(error, interval[0], interval[1], **kwargs)[0]
        else:
            def error(x):
                return functional(pdf1(x, i), pdf2(x, i))
            return quadrature(error, interval[0], interval[1], **kwargs)[0]



#! /usr/bin/env python

# Copyright (C) 2014-2020 The BET Development Team

r"""
This example generates samples on a 2D grid and evaluates
a nonlinear map to a 1d or 2d space. Modify QoI_num in myModel.py
to change the dimension of the output.
The maps are defined
as quantities of interest (QoI) defined as spatial
observations of the solution to the elliptic PDE .. math::
  :nowrap:

  \begin{cases}
    -\nabla \cdot (A(\lambda)\cdot\nabla u) &= f(x,y;\lambda), \ (x,y)\in\Omega, \\
    u|_{\partial \Omega} &= 0,
  \end{cases}

where :math:`A(\lambda)=\text{diag}(1/\lambda_1^2,1/\lambda_2^2)`,
:math: `f(x,y;\lambda) = \pi^2 \sin(\pi x\lambda_1)\sin(\pi y \lambda_2)`,
and :math:`\Omega=[0,1]\times[0,1]`.

Probabilities
in the parameter space are calculated using emulated points.
1D and 2D marginals are calculated, smoothed, and plotted.
"""


import numpy as np
import bet.calculateP.simpleFunP as simpleFunP
import bet.calculateP.calculateP as calculateP
import bet.postProcess.plotP as plotP
import bet.postProcess.plotDomains as plotD
import bet.sample as samp
import bet.sampling.basicSampling as bsam

from myModel import my_model

# Define the sampler that will be used to create the discretization
# object, which is the fundamental object used by BET to compute
# solutions to the stochastic inverse problem.
# The sampler and my_model is the interface of BET to the model,
# and it allows BET to create input/output samples of the model.
sampler = bsam.sampler(my_model)

# Initialize 2-dimensional input parameter sample set object
input_samples = samp.sample_set(2)

# Set parameter domain
input_samples.set_domain(np.array([[3.0, 6.0],
                                   [1.0, 5.0]]))


'''
Suggested changes for user:

Try with and without random sampling.

If using random sampling, try num_samples = 1E3 and 1E4.
What happens when num_samples = 1E2?
Try using 'lhs' instead of 'random' in the random_sample_set.

If using regular sampling, try different numbers of samples
per dimension.
'''
# Generate samples on the parameter space
randomSampling = True
if randomSampling is True:
    input_samples = sampler.random_sample_set(
        'uniform',
        input_samples, num_samples=1E4)
else:
    input_samples = sampler.regular_sample_set(
        input_samples, num_samples_per_dim=[50, 50])

'''
Suggested changes for user:

A standard Monte Carlo (MC) assumption is that every Voronoi cell
has the same volume. If a regular grid of samples was used, then
the standard MC assumption is true.

See what happens if the MC assumption is not assumed to be true, and
if different numbers of points are used to estimate the volumes of
the Voronoi cells.
'''
MC_assumption = True
# Estimate volumes of Voronoi cells associated with the parameter samples
if MC_assumption is False:
    input_samples.estimate_volume(n_mc_points=1E5)
else:
    input_samples.estimate_volume_mc()

# Create the discretization object using the input samples
my_discretization = sampler.compute_qoi_and_create_discretization(input_samples,
                                                                  savefile='NonlinearExample.txt.gz')

'''
Suggested changes for user:

Try different reference parameters.
'''
# Define the reference parameter
param_ref = np.array([5.5, 4.5])
#param_ref = np.array([4.5, 3.0])
#param_ref = np.array([3.5, 1.5])

# Compute the reference QoI
Q_ref = my_model(param_ref)

# Create some plots of input and output discretizations
plotD.scatter_2D(input_samples, ref_sample=param_ref,
                 filename='nonlinearMapParameterSamples',
                 file_extension='.png')
if Q_ref.size == 2:
    plotD.show_data_domain_2D(
        my_discretization,
        Q_ref=Q_ref,
        file_extension=".png")

'''
Suggested changes for user:

Try different ways of discretizing the probability measure on D defined
as a uniform probability measure on a rectangle or interval depending
on choice of QoI_num in myModel.py.
'''
randomDataDiscretization = False
if randomDataDiscretization is False:
    simpleFunP.regular_partition_uniform_distribution_rectangle_scaled(
        data_set=my_discretization, Q_ref=Q_ref, rect_scale=0.25,
        cells_per_dimension=3)
else:
    simpleFunP.uniform_partition_uniform_distribution_rectangle_scaled(
        data_set=my_discretization, Q_ref=Q_ref, rect_scale=0.25,
        M=50, num_d_emulate=1E5)

# calculate probabilities
calculateP.prob(my_discretization)

########################################
# Post-process the results
########################################
'''
Suggested changes for user:

At this point, the only thing that should change in the plotP.* inputs
should be either the nbins values or sigma (which influences the kernel
density estimation with smaller values implying a density estimate that
looks more like a histogram and larger values smoothing out the values
more).

There are ways to determine "optimal" smoothing parameters (e.g., see CV, GCV,
and other similar methods), but we have not incorporated these into the code
as lower-dimensional marginal plots generally have limited value in understanding
the structure of a high dimensional non-parametric probability measure.
'''
# calculate 2d marginal probs
(bins, marginals2D) = plotP.calculate_2D_marginal_probs(input_samples,
                                                        nbins=[20, 20])
# smooth 2d marginals probs (optional)
marginals2D = plotP.smooth_marginals_2D(marginals2D, bins, sigma=0.5)

# plot 2d marginals probs
plotP.plot_2D_marginal_probs(marginals2D, bins, input_samples, filename="nomlinearMap",
                             lam_ref=param_ref, file_extension=".png", plot_surface=False)

# calculate 1d marginal probs
(bins, marginals1D) = plotP.calculate_1D_marginal_probs(input_samples,
                                                        nbins=[20, 20])
# smooth 1d marginal probs (optional)
marginals1D = plotP.smooth_marginals_1D(marginals1D, bins, sigma=0.5)
# plot 2d marginal probs
plotP.plot_1D_marginal_probs(marginals1D, bins, input_samples, filename="nonlinearMap",
                             lam_ref=param_ref, file_extension=".png")



# Copyright (C) 2014-2020 The BET Development Team

"""
This module contains unittests for :mod:`~bet.sampling.basicSampling:`
"""

import unittest
import os
import bet
import numpy.testing as nptest
import numpy as np
import scipy.io as sio
import bet.sampling.LpGeneralizedSamples as lp


def test_Lp_generalized_normal():
    """
    Tests :meth:`bet.Lp_generalized_samples.Lp_generalized_normal`

    This test only verifies the mean, but not the variance.

    """
    # 1D
    nptest.assert_allclose(np.mean(lp.Lp_generalized_normal(1, 1000), 0),
                           np.zeros((1,)), atol=1e-1)
    # 2D
    nptest.assert_allclose(np.mean(lp.Lp_generalized_normal(2, 1000), 0),
                           np.zeros((2,)), atol=1e-1)
    # 3D
    nptest.assert_allclose(np.mean(lp.Lp_generalized_normal(3, 1000), 0),
                           np.zeros((3,)), atol=1e-1)


def verify_norm_and_mean(x, r, p):
    """

    Verify that all of the samples in `x` are within the Lp ball centered at 0.
    Verify the mean of `x` is zero.

    :param x: Array containing a set of samples
    :type x: :class:`numpy.ndarry` of shape (num, dim)
    :param float r: radius of the Lp ball
    :param float p: 0 < p <= infinity, p of the Lp ball

    """
    if np.isinf(p):
        xpnorm = np.max(np.abs(x), 1)
    else:
        xpnorm = np.sum(np.abs(x)**p, 1)**(1. / p)
    assert np.all(xpnorm <= r)
    nptest.assert_allclose(np.mean(x, 0), np.zeros((x.shape[1],)), atol=1e-1)


def verify_norm(x, r, p):
    """

    Verify that all of the samples in `x` are within the Lp ball centered at 0.

    :param x: Array containing a set of samples
    :type x: :class:`numpy.ndarry` of shape (num, dim)
    :param float r: radius of the Lp ball
    :param float p: 0 < p <= infinity, p of the Lp ball

    """
    if np.isinf(p):
        xpnorm = np.max(np.abs(x), 1)
    else:
        xpnorm = np.sum(np.abs(x)**p, 1)**(1. / p)
    assert np.all(xpnorm <= r)


def test_Lp_generalized_uniform():
    """
    Tests :meth:`bet.Lp_generalized_samples.Lp_generalized_uniform`

    This test only verifies the mean, but not the variance.

    """
    # 1D
    x = lp.Lp_generalized_uniform(1, 1000)
    nptest.assert_allclose(np.mean(x, 0), np.zeros((1,)), atol=1e-1)
    assert np.all(np.logical_and(x <= 1., x >= -1))

    # 2D
    p = 1
    x = lp.Lp_generalized_uniform(2, 1000, p)
    verify_norm_and_mean(x, 1.0, p)

    p = 2
    x = lp.Lp_generalized_uniform(2, 1000, p)
    verify_norm_and_mean(x, 1.0, p)

    p = 3
    x = lp.Lp_generalized_uniform(2, 1000, p)
    verify_norm_and_mean(x, 1.0, p)

    p = np.inf
    x = lp.Lp_generalized_uniform(2, 1000, p)
    verify_norm_and_mean(x, 1.0, p)

    # 3D
    p = 1
    x = lp.Lp_generalized_uniform(3, 1000, p)
    verify_norm_and_mean(x, 1.0, p)

    p = 2
    x = lp.Lp_generalized_uniform(3, 1000, p)
    verify_norm_and_mean(x, 1.0, p)

    p = 3
    x = lp.Lp_generalized_uniform(3, 1000, p)
    verify_norm_and_mean(x, 1.0, p)

    p = np.inf
    x = lp.Lp_generalized_uniform(3, 1000, p)
    verify_norm_and_mean(x, 1.0, p)


def test_Lp_generalized_beta():
    """
    Tests :meth:`bet.Lp_generalized_samples.Lp_generalized_beta`

    This test only verifies the mean, but not the variance.

    """
    # 1D
    x = lp.Lp_generalized_beta(1, 1000)
    nptest.assert_allclose(np.mean(x, 0), np.zeros((1,)), atol=1e-1)
    assert np.all(np.logical_and(x <= 1., x >= -1))

    # 2D
    p = 1
    x = lp.Lp_generalized_beta(2, 1000, p)
    verify_norm(x, 1.0, p)

    p = 2
    x = lp.Lp_generalized_beta(2, 1000, p)
    verify_norm_and_mean(x, 1.0, p)

    p = 3
    x = lp.Lp_generalized_beta(2, 1000, p)
    verify_norm(x, 1.0, p)

    # 3D
    p = 1
    x = lp.Lp_generalized_beta(3, 1000, p)
    verify_norm(x, 1.0, p)

    p = 2
    x = lp.Lp_generalized_beta(3, 1000, p)
    verify_norm_and_mean(x, 1.0, p)

    p = 3
    x = lp.Lp_generalized_beta(3, 1000, p)
    verify_norm(x, 1.0, p)



# Copyright (C) 2014-2020 The BET Development Team

r"""
This subpackage provides classes and methods for calculating the
probability measure :math:`P_{\Lambda}`.

* :mod:`~bet.calculateP.calculateP` provides methods for approximating probability densities in the measure-based approach.
* :mod:`~bet.calculateP.simpleFunP` provides methods for creating simple function approximations of probability densities for the measure-based approach.
* :mod:`~bet.calculateP.calculateR` provides methods for density-based approach.
* :mod:`~bet.calculateP.calculateError` provides methods for approximating numerical and sampling errors.
"""
__all__ = ['calculateP', 'simpleFunP', 'calculateError', 'calculateR']



# Copyright (C) 2014-2020 The BET Development Team

r"""
This module provides methods for calculating error estimates of
the probability measure for calculate probability measures. See
`Butler et al. 2015. <http://arxiv.org/pdf/1407.3851>`.

* :mod:`~bet.calculateErrors.cell_connectivity_exact` calculates the connectivity of cells.
* :mod:`~bet.calculateErrors.boundary_sets` calculates which cells are on the boundary and strictly interior for contour events.
* :class:`~bet.calculateErrors.sampling_error` is for calculating error estimates due to sampling.
* :class:`~bet.calculateErrors.model_error` is for calculating error estimates due to error in solution of QoIs

"""

import logging
import numpy as np
from bet.Comm import comm, MPI
import bet.sample as samp


class wrong_argument_type(Exception):
    """
    Exception for when the argument is not one of the acceptible
    types.
    """


def cell_connectivity_exact(disc):
    """

    Calculates contour events of the cells and its neighbors.

    :param disc: An object containing the discretization information.
    :type disc: :class:`bet.sample.discretization`

    :rtype: list
    :returns: list of lists of neighboring cells

    """
    from scipy.spatial import Delaunay
    from collections import defaultdict
    import itertools
    import numpy.linalg as nlinalg

    # Check inputs
    if not isinstance(disc, samp.discretization):
        msg = "The argument must be of type bet.sample.discretization."
        raise wrong_argument_type(msg)

    if not isinstance(disc.get_input_sample_set(), samp.voronoi_sample_set):
        msg = "disc._input_sample_set must be of type bet.sample.voronoi"
        msg += "_sample_set defined with the 2-norm"
        raise wrong_argument_type(msg)
    elif disc._input_sample_set._p_norm != 2.0:
        msg = "disc._input_sample_set must be of type bet.sample.voronoi"
        msg += "_sample_set defined with the 2-norm"
        raise wrong_argument_type(msg)

    num = disc.check_nums()
    # Set up necessary pointers
    if disc.get_io_ptr() is None:
        disc.set_io_ptr()

    if disc._input_sample_set._dim == 1:
        # Adding contours on the left
        s_sort = disc._input_sample_set._values.flat[:].argsort()
        neiList = defaultdict(set)
        for p in range(num):
            order = s_sort[p]
            if order > 0:
                val = np.equal(s_sort, order - 1)
                args = np.argwhere(val)
                neiList[p].add(disc._io_ptr[args[0][0]])
                neiList[args[0][0]].add(disc._io_ptr[p])
    else:
        # Form Delaunay triangulation
        tri = Delaunay(disc._input_sample_set._values)

        # Find neighbors
        neiList = defaultdict(set)
        for p in tri.vertices:
            for i, j in itertools.combinations(p, 2):
                neiList[i].add(disc._io_ptr[j])
                neiList[j].add(disc._io_ptr[i])
    # Get rid of redundant entries
    for i in range(num):
        neiList[i] = list(set(neiList[i]))

    return neiList


def boundary_sets(disc, nei_list):
    """

    Calculates the the neighboring Voronoi cells for each cell.

    :param disc: An object containing the discretization information.
    :type disc: :class:`bet.sample.discretization`
    :param nei_list: list of lists defining contour events of neighboring cells.
    :type nei_list: list

    :rtype: tuple
    :returns: (:math:`B_N, C_N`) where B_N are the cells strictly on the
        interior of a contour event and C_N are the cells on the boundary
        of a contour eventas defined in
        `Butler et al. 2015. <http://arxiv.org/pdf/1407.3851>`


    """
    from collections import defaultdict

    # Check inputs
    if not isinstance(disc, samp.discretization):
        msg = "The argument must be of type bet.sample.discretization."
        raise wrong_argument_type(msg)

    if not isinstance(disc._input_sample_set, samp.voronoi_sample_set):
        msg = "disc._input_sample_set must be of type bet.sample.voronoi"
        msg += "_sample_set defined with the 2-norm"
        raise wrong_argument_type(msg)
    elif disc._input_sample_set._p_norm != 2.0:

        msg = "disc._input_sample_set must be of type bet.sample.voronoii"
        msg += "_sample_set defined with the 2-norm"
        raise wrong_argument_type(msg)

    num = disc.check_nums()

    # Form necessary pointers
    if disc.get_io_ptr() is None:
        disc.set_io_ptr()
    # Define strictly interior and boundary cells for each contour event
    B_N = defaultdict(list)
    C_N = defaultdict(list)
    for i in range(num):
        contour_event = disc._io_ptr[i]
        if nei_list[i] == [contour_event]:
            B_N[contour_event].append(i)
        for j in nei_list[i]:
            C_N[j].append(i)

    return (B_N, C_N)


class sampling_error(object):
    """
    A class for calculating the error due to sampling for a discretization.
    """

    def __init__(self, disc, exact=True):
        """

        Set things up for a given discretization

        :param disc: An object containing the discretization information.
        :type disc: :class:`bet.sample.discretization`
        :param exact: Whether or not to use exact connectivity
        :type exact: bool

        """
        # Check inputs
        if not isinstance(disc, samp.discretization):
            msg = "The argument must be of type bet.sample.discretization."
            raise wrong_argument_type(msg)

        #: :class:`bet.sample.discretization` that defines the problem
        self.disc = disc
        #: number of inputs and outputs
        self.num = self.disc.check_nums()

        # Set up neighbor list and B_N and C_N
        if exact:
            nei_list = cell_connectivity_exact(self.disc)
        else:
            msg = "sampling_error only handles exact connectivity for now."
            raise NotImplementedError(msg)
        #: dictionaries of interior and boundary sets
        (self.B_N, self.C_N) = boundary_sets(self.disc, nei_list)

    def calculate_for_contour_events(self):
        """

        Calculate the sampling error bounds for each contour event.  If volumes
        are already calculated these are used. If not, volume emulation is used
        if an emulated input set exists.  Otherwise the MC assumption is
        made.


        :rtype: tuple
        :returns: (``up_list``, ``low_list``) where ``up_list`` is a list of
            the upper bounds for each contour event and ``low_list`` is a list
            of the lower bounds.

        """
        up_list = []
        low_list = []
        # Check for and possibly calculate volumes
        if self.disc._input_sample_set._volumes is None:
            if self.disc._emulated_input_sample_set is not None:
                logging.warning("Using emulated points to estimate volumes.")
                self.disc._input_sample_set.estimate_volume_emulated(self.
                                                                     disc._emulated_input_sample_set)
            else:
                logging.warning("Making MC assumption to estimate volumes.")
                self.disc._input_sample_set.estimate_volume_mc()

        # Loop over contour events and calculate error bounds
        ops_num = self.disc._output_probability_set.check_num()
        for i in range(ops_num):
            # contour event :math:`\mathcal{A} = A_{i,N}`
            if self.disc._output_probability_set._probabilities[i] > 0.0:
                lam_vol = np.zeros((self.num,))
                indices = np.equal(self.disc._io_ptr, i)
                lam_vol[indices] = self.disc._input_sample_set._volumes[indices]
                if i in self.B_N:
                    # val1 = :math:`\mu_{\Lambda}(B_{i,N})`
                    val1 = np.sum(self.disc._input_sample_set._volumes[self.
                                                                       B_N[i]])
                    # val2 = :math:`\mu_{\Lambda}(\mathcal{A} \cap B_{i,N})`
                    val2 = np.sum(lam_vol[self.B_N[i]])
                    # val3 = :math:`\mu_\Lambda(C_{i,N})`
                    val3 = np.sum(self.disc._input_sample_set._volumes[self.
                                                                       C_N[i]])
                    # val4 = :math:`\mu_{\Lambda}(\mathcal{A} \cap C_{i,N})`
                    val4 = np.sum(lam_vol[self.C_N[i]])

                    term1 = val2 / val3 - 1.0
                    term2 = val4 / val1 - 1.0
                    up_list.append(self.disc._output_probability_set.
                                   _probabilities[i] * max(term1, term2))
                    low_list.append(self.disc._output_probability_set.
                                    _probabilities[i] * min(term1, term2))
                else:
                    up_list.append(float('nan'))
                    low_list.append(float('nan'))

            else:
                up_list.append(0.0)
                low_list.append(0.0)

        return (up_list, low_list)

    def calculate_for_sample_set_region(self, s_set,
                                        region, emulated_set=None):
        r"""
        Calculate the sampling error bounds for a region of the input space
        defined by a sample set object which defines an event :math:`A`.

        :param s_set: sample set for which to calculate error
        :type s_set: :class:`bet.sample.sample_set_base`
        :param int region: region of s_set for which to calculate error
        :param emulated_set: sample set for volume emulation
        :type emulated_set: :class:`bet.sample_set_base`

        :rtype: tuple
        :returns: (``upper_bound``, ``lower_bound``) the upper and lower bounds
            for the error.

        """
        # Set up marker
        self.disc._input_sample_set.local_to_global()
        if s_set._region is None:
            msg = "regions must be defined for the sample set."
            raise wrong_argument_type(msg)
        marker = np.equal(s_set._region, region)
        if not np.any(marker):
            msg = "The given region does not exist."
            raise wrong_argument_type(msg)

        # Set up discretizations
        if emulated_set is not None:
            disc = self.disc.copy()
            disc.set_emulated_input_sample_set(emulated_set)
            disc.set_emulated_ii_ptr(globalize=False)
            disc_new = samp.discretization(input_sample_set=s_set,
                                           output_sample_set=s_set,
                                           emulated_input_sample_set=emulated_set)
            disc_new.set_emulated_ii_ptr(globalize=False)
        elif self.disc._emulated_input_sample_set is not None:
            msg = "Using emulated_input_sample_set for volume emulation"
            logging.warning(msg)
            disc = self.disc
            if disc._emulated_ii_ptr is None:
                disc.set_emulated_ii_ptr(globalize=False)
            disc_new = samp.discretization(input_sample_set=s_set,
                                           output_sample_set=s_set,
                                           emulated_input_sample_set=self.
                                           disc.
                                           _emulated_input_sample_set)

            disc_new.set_emulated_ii_ptr(globalize=False)
        else:
            logging.warning("Using MC assumption for calculating volumes.")
            disc = self.disc.copy()
            disc.set_emulated_input_sample_set(disc._input_sample_set)
            disc.set_emulated_ii_ptr(globalize=False)

            disc_new = samp.discretization(input_sample_set=s_set,
                                           output_sample_set=s_set,
                                           emulated_input_sample_set=self.
                                           disc._input_sample_set)
            disc_new.set_emulated_ii_ptr(globalize=False)

        # Emulated points in the the region
        in_A = marker[disc_new._emulated_ii_ptr_local]

        upper_bound = 0.0
        lower_bound = 0.0
        # Loop over contour intervals and add error contributions
        ops_num = self.disc._output_probability_set.check_num()
        for i in range(ops_num):
            # Contribution from contour event :math:`A_{i,N}`
            if self.disc._output_probability_set._probabilities[i] > 0.0:
                indices = np.equal(disc._io_ptr, i)
                in_Ai = indices[disc._emulated_ii_ptr_local]
                # sum1 :math:`\mu_{\Lambda}(A \cap A_{i,N})`
                sum1 = np.sum(np.logical_and(in_A, in_Ai))
                # sum2 :math:`\mu_{\Lambda}(A_{i,N})`
                sum2 = np.sum(in_Ai)
                sum1 = comm.allreduce(sum1, op=MPI.SUM)
                sum2 = comm.allreduce(sum2, op=MPI.SUM)
                if sum2 == 0.0:
                    return (float('nan'), float('nan'))
                E = float(sum1) / float(sum2)

                in_B_N = np.zeros(in_A.shape, dtype=np.bool)
                for j in self.B_N[i]:
                    in_B_N = np.logical_or(np.equal(disc.
                                                    _emulated_ii_ptr_local, j), in_B_N)

                in_C_N = np.zeros(in_A.shape, dtype=np.bool)
                for j in self.C_N[i]:
                    in_C_N = np.logical_or(np.equal(disc.
                                                    _emulated_ii_ptr_local, j), in_C_N)
                # sum3 :math:`\mu_{\Lambda}(A \cap B_N)`
                sum3 = np.sum(np.logical_and(in_A, in_B_N))
                # sum4 :math:`\mu_{\Lambda}(C_N)`
                sum4 = np.sum(in_C_N)
                sum3 = comm.allreduce(sum3, op=MPI.SUM)
                sum4 = comm.allreduce(sum4, op=MPI.SUM)
                if sum4 == 0.0:
                    return (float('nan'), float('nan'))
                term1 = float(sum3) / float(sum4) - E

                # sum5 :math:`\mu_{\Lambda}(A \cap C_N)`
                sum5 = np.sum(np.logical_and(in_A, in_C_N))
                # sum6 :math:`\mu_{\Lambda}(B_N)`
                sum6 = np.sum(in_B_N)
                sum5 = comm.allreduce(sum5, op=MPI.SUM)
                sum6 = comm.allreduce(sum6, op=MPI.SUM)
                if sum6 == 0.0:
                    return (float('nan'), float('nan'))
                term2 = float(sum5) / float(sum6) - E

                upper_bound += self.disc._output_probability_set.\
                    _probabilities[i] * max(term1, term2)
                lower_bound += self.disc._output_probability_set.\
                    _probabilities[i] * min(term1, term2)
        return (upper_bound, lower_bound)


class model_error(object):
    """
    A class for calculating the error due to numerical error
    for a discretization.
    """

    def __init__(self, disc):
        """

        Set things up for a given discretization

        :param disc: An object containing the discretization information.
        :type disc: :class:`bet.sample.discretization`

        """
        # Check inputs
        if not isinstance(disc, samp.discretization):
            msg = "The argument must be of type bet.sample.discretization."
            raise wrong_argument_type(msg)
        disc._output_sample_set.global_to_local()
        #: :class:`bet.sample.discretiztion` defining the problem
        self.disc = disc
        # discretization must have error estimates
        if self.disc._output_sample_set._error_estimates is None:
            if self.disc._output_sample_set._error_estimates_local is None:
                msg = "Error estimates for the output sample set are required."
                raise wrong_argument_type(msg)

        #: number of inputs and outputs
        self.num = self.disc.check_nums()
        if self.disc._io_ptr_local is None:
            self.disc.set_io_ptr()

        # Setup new discretization object adding error estimates
        #: :class:`bet.sample.discretiztion` from adding error estimates
        self.disc_new = disc.copy()
        self.disc_new._output_sample_set._values_local += self.disc.\
            _output_sample_set._error_estimates_local
        self.disc_new.set_io_ptr(globalize=False)
        self.disc_new._io_ptr = None

    def calculate_for_contour_events(self):
        r"""

        Calculate the numerical error for each contour event.

        :rtype: list
        :returns: ``er_list``, a list of the error estimates
            for each contour event.

        """
        # Calculate volumes if necessary
        if self.disc._input_sample_set._volumes is None:
            if self.disc._emulated_input_sample_set is not None:
                logging.warning("Using emulated points to estimate volumes.")
                self.disc._input_sample_set.estimate_volume_emulated(self.
                                                                     disc._emulated_input_sample_set)
            else:
                logging.warning("Making MC assumption to estimate volumes.")
                self.disc._input_sample_set.estimate_volume_mc()
        # Localize if necessary
        if self.disc._input_sample_set._volumes_local is None:
            self.disc._input_sample_set.global_to_local()

        # Loop over contour events and add contributions
        er_list = []
        ops_num = self.disc._output_probability_set.check_num()
        for i in range(ops_num):
            if self.disc._output_probability_set._probabilities[i] > 0.0:
                # JiA, Ji, Jie, and JiAe are defined ast in
                # `Butler et al. 2015. <http://arxiv.org/pdf/1407.3851>`
                ind1 = np.equal(self.disc._io_ptr_local, i)
                ind2 = np.equal(self.disc_new._io_ptr_local, i)
                JiA = np.sum(self.disc._input_sample_set._volumes_local[ind1])
                Ji = JiA
                JiAe = np.sum(self.disc._input_sample_set._volumes_local[
                    np.logical_and(ind1, ind2)])
                Jie = np.sum(self.disc._input_sample_set._volumes_local[ind2])

                JiA = comm.allreduce(JiA, op=MPI.SUM)
                Ji = comm.allreduce(Ji, op=MPI.SUM)
                JiAe = comm.allreduce(JiAe, op=MPI.SUM)
                Jie = comm.allreduce(Jie, op=MPI.SUM)
                er_list.append(self.disc._output_probability_set.
                               _probabilities[i] * ((JiA * Jie - JiAe * Ji) / (Ji * Jie)))
            else:
                er_list.append(0.0)

        return er_list

    def calculate_for_sample_set_region(self, s_set,
                                        region, emulated_set=None):
        """
        Calculate the numerical error estimate for a region of the input space
        defined by a sample set object.

        :param s_set: sample set for which to calculate error
        :type s_set: :class:`bet.sample.sample_set_base`
        :param region: region of s_set for which to calculate error
        :type region: int
        :param emulated_set: sample set for volume emulation
        :type emulated_sample_set: :class:`bet.sample_set_base`

        :rtype: float
        :returns: ``er_est``, the numerical error estimate for the region

        """
        # Set up marker
        if s_set._region is None:
            msg = "regions must be defined for the sample set."
            raise wrong_argument_type(msg)
        marker = np.equal(s_set._region, region)
        if not np.any(marker):
            msg = "The given region does not exist."
            raise wrong_argument_type(msg)

        # Setup discretizations
        if emulated_set is not None:
            self.disc._input_sample_set.local_to_global()
            self.disc.globalize_ptrs()
            self.disc_new.globalize_ptrs()

            disc = self.disc.copy()
            disc.set_emulated_input_sample_set(emulated_set)
            disc.set_emulated_ii_ptr(globalize=False)

            disc_new_set = samp.discretization(input_sample_set=s_set,
                                               output_sample_set=s_set,
                                               emulated_input_sample_set=emulated_set)
            disc_new_set.set_emulated_ii_ptr(globalize=False)
        elif self.disc._emulated_input_sample_set is not None:
            self.disc._input_sample_set.local_to_global()
            msg = "Using emulated_input_sample_set for volume emulation"
            logging.warning(msg)
            self.disc.globalize_ptrs()
            self.disc_new.globalize_ptrs()
            disc = self.disc
            if disc._emulated_ii_ptr_local is None:
                disc.set_emulated_ii_ptr(globalize=False)
            self.disc_new.set_emulated_ii_ptr(globalize=False)
            disc_new_set = samp.discretization(input_sample_set=s_set,
                                               output_sample_set=s_set, emulated_input_sample_set=disc._emulated_input_sample_set)
            disc_new_set.set_emulated_ii_ptr(globalize=False)
        else:
            logging.warning("Using MC assumption for volumes.")
            return self.calculate_for_sample_set_region_mc(s_set, region)

        # Setup pointers
        ptr1 = disc._emulated_ii_ptr_local
        ptr3 = disc_new_set._emulated_ii_ptr_local

        # Check if in the region
        in_A = marker[ptr3]

        # Loop over contour events and add error contribution
        er_est = 0.0
        ops_num = self.disc._output_probability_set.check_num()

        for i in range(ops_num):
            if self.disc._output_probability_set._probabilities[i] > 0.0:
                # JiA, Ji, Jie, and JiAe are defined ast in
                # `Butler et al. 2015. <http://arxiv.org/pdf/1407.3851>`_
                indices1 = np.equal(self.disc._io_ptr, i)
                in_Ai1 = indices1[ptr1]
                indices2 = np.equal(self.disc_new._io_ptr, i)
                in_Ai2 = indices2[ptr1]
                JiA_local = float(np.sum(np.logical_and(in_A, in_Ai1)))
                JiA = comm.allreduce(JiA_local, op=MPI.SUM)
                Ji_local = float(np.sum(in_Ai1))
                Ji = comm.allreduce(Ji_local, op=MPI.SUM)
                JiAe_local = float(np.sum(np.logical_and(in_A, in_Ai2)))
                JiAe = comm.allreduce(JiAe_local, op=MPI.SUM)
                Jie_local = float(np.sum(in_Ai2))
                Jie = comm.allreduce(Jie_local, op=MPI.SUM)
                er_est += self.disc._output_probability_set._probabilities[i]\
                    * ((JiA * Jie - JiAe * Ji) / (Ji * Jie))

        return er_est

    def calculate_for_sample_set_region_mc(self, s_set,
                                           region):
        """
        Calculate the numerical error estimate for a region of the input space
        defined by a sample set object, using the MC assumption.

        :param s_set: sample set for which to calculate error
        :type s_set: :class:`bet.sample.sample_set_base`
        :param region: region of s_set for which to calculate error
        :type region: int

        :rtype float
        :returns: ``er_est``, the numerical error estimate for the region

        """
        # Set up marker
        if s_set._region is None:
            msg = "regions must be defined for the sample set."
            raise wrong_argument_type(msg)
        marker = np.equal(s_set._region, region)
        if not np.any(marker):
            msg = "The given region does not exist."
            raise wrong_argument_type(msg)

        disc_new_set = samp.discretization(input_sample_set=s_set,
                                           output_sample_set=s_set,
                                           emulated_input_sample_set=self.disc._input_sample_set)
        disc_new_set.set_emulated_ii_ptr(globalize=False)
        # Check if in the region
        in_A = marker[disc_new_set._emulated_ii_ptr_local]

        # Loop over contour events and add error contribution
        er_est = 0.0
        ops_num = self.disc._output_probability_set.check_num()
        num_local = self.disc._input_sample_set.check_num_local()
        self.disc._input_sample_set._error_id_local = np.zeros((num_local,))

        for i in range(ops_num):
            if self.disc._output_probability_set._probabilities[i] > 0.0:
                # JiA, Ji, Jie, and JiAe are defined ast in
                # `Butler et al. 2015. <http://arxiv.org/pdf/1407.3851>`
                in_Ai1 = np.equal(self.disc._io_ptr_local, i)
                in_Ai2 = np.equal(self.disc_new._io_ptr_local, i)
                JiA_local = float(np.sum(np.logical_and(in_A, in_Ai1)))
                JiA = comm.allreduce(JiA_local, op=MPI.SUM)
                Ji_local = float(np.sum(in_Ai1))
                Ji = comm.allreduce(Ji_local, op=MPI.SUM)
                JiAe_local = float(np.sum(np.logical_and(in_A, in_Ai2)))
                JiAe = comm.allreduce(JiAe_local, op=MPI.SUM)
                Jie_local = float(np.sum(in_Ai2))
                Jie = comm.allreduce(Jie_local, op=MPI.SUM)
                if Ji * Jie == 0:
                    er_cont = np.inf
                else:
                    er_cont = self.disc._output_probability_set._probabilities[i]\
                        * ((JiA * Jie - JiAe * Ji) / (Ji * Jie))
                er_est += er_cont
                error_cells1 = np.logical_and(np.logical_and(in_Ai1,
                                                             np.logical_not(in_A)), np.logical_and(in_Ai2, in_A))
                error_cells2 = np.logical_and(np.logical_and(in_Ai2,
                                                             np.logical_not(in_A)), np.logical_and(in_Ai1, in_A))
                error_cells3 = np.not_equal(in_Ai1, in_Ai2)
                error_cells = np.logical_or(error_cells1, error_cells2)
                error_cells = np.logical_or(error_cells, error_cells3)

                error_cells_num_local = float(np.sum(error_cells))
                error_cells_num = comm.allreduce(error_cells_num_local,
                                                 op=MPI.SUM)
                if error_cells_num != 0:
                    self.disc._input_sample_set._error_id_local[error_cells] \
                        += er_cont / error_cells_num
        return er_est



# Copyright (C) 2014-2020 The BET Development Team

"""
Here, we consider a simple example where a parameter space is given
by a 5-dimensional hypercube and the goal is to choose an optimal QoI
map from a space of possible QoI maps, denoted by :math:`\mathcal{Q}`,
where each QoI map is linear.
We use this simple example to demonstrate the use of the code to
optimally choose which possible QoI map does the best job of "scaling"
inverse sets to smaller sets.
The idea is that if we generally consider a set of high probability
in a particular data space defined by the range of a QoI map, we would
prefer that the inverse of this set is as small as possible in order to
try and identify the parameter responsible for the data.
This only makes sense for stochastic inverse problems framed within the
context of parameter identification under uncertainty.
In other words, when the problem is that the data are uncertain due to
measurement uncertainty and there is a true/exact parameter responsible
for whichever uncertain data is observed, then this is the type of problem
for which this optimization criteria is most appropriate.

This example generates uniform random samples in the unit hypercube and
corresponding QoIs (data) generated by a linear map Q.
We then calculate the gradients using an RBF scheme and use
the gradient information to choose the optimal set of
2 (3, 4, ... input_dim) QoIs to use in the inverse problem.
"""

import numpy as np
import bet.sensitivity.gradients as grad
import bet.sensitivity.chooseQoIs as cqoi
import bet.calculateP.simpleFunP as simpleFunP
import bet.calculateP.calculateP as calculateP
import bet.postProcess.postTools as postTools
import bet.Comm as comm
import bet.sample as sample

# Set up the info for the spaces
input_dim = 5
QoI_scalar_map_num = 10
num_samples = 1E5
num_centers = 10

# Let the map Q be a random matrix of size (QoI_scalar_map_num, input_dim)
np.random.seed(0)
Q = np.random.random([QoI_scalar_map_num, input_dim])

# Initialize some sample objects we will need
input_samples = sample.sample_set(input_dim)
output_samples = sample.sample_set(QoI_scalar_map_num)

# Choose random samples in parameter space to solve the model
input_samples.set_values(np.random.uniform(
    0, 1, [np.int(num_samples), input_dim]))


# Compute the output values with the map Q
output_samples.set_values(Q.dot(input_samples.get_values().transpose()).
                          transpose())

# Calculate the gradient vectors at some subset of the samples.  Here the
# *normalize* argument is set to *True* because we are using *bin_ratio* to
# determine the uncertainty in our data.
cluster_discretization = sample.discretization(input_samples, output_samples)
# We will approximate the jacobian at each of the centers
center_discretization = grad.calculate_gradients_rbf(cluster_discretization,
                                                     num_centers, normalize=True)

# With these gradient vectors, we are now ready to choose an optimal set of
# QoIs to use in the inverse problem, based on minimizing the support of the
# inverse solution (measure).  The most robust method for this is
# :meth:~bet.sensitivity.chooseQoIs.chooseOptQoIs_large which returns the
# best set of 2, 3, 4 ... until input_dim.  This method returns a list of
# matrices.  Each matrix has 10 rows, the first column representing the
# expected inverse measure ratio, and the rest of the columns the corresponding
# QoI indices.
input_samples_center = center_discretization.get_input_sample_set()
best_sets = cqoi.chooseOptQoIs_large(input_samples_center, measure=True)

###############################################################################

# At this point we have determined the optimal set of QoIs to use in the inverse
# problem.  Now we compare the support of the inverse solution using
# different sets of these QoIs.  We set Q_ref to correspond to the center of
# the parameter space.  We choose the set of QoIs to consider.

QoI_indices = best_sets[0][0, 1:].astype(
    int)  # Chooses the optimal set of 2 QoI
# QoI_indices = best_sets[0][1,1:].astype(int) # Chooses the second most optimal set of 2 QoI
# QoI_indices = best_sets[0][9,1:].astype(int) # Chooses the least optimal set of 2 QoI
# QoI_indices = best_sets[3][0,1:].astype(int) # Chooses the optimal set of 5 QoI
# QoI_indices = best_sets[3][1,1:].astype(int) # Chooses the second most
# optimal set of 5 QoI

# Setup the output sample set associated to the QoI choice
output_samples._dim = len(QoI_indices)
output_samples.set_values(output_samples.get_values()[:, QoI_indices])

# Define the reference point in the output space to correspond to the center of
# the input space.
param_ref = 0.5 * np.ones(input_dim)
Q_ref = Q[QoI_indices, :].dot(param_ref)

# Define the level of uncertainty in the measured reference datum
rect_scale = 0.25

# Make the MC assumption and compute the volumes of each voronoi cell
input_samples.estimate_volume_mc()

# Create discretization object
my_discretization = sample.discretization(input_sample_set=input_samples,
                                          output_sample_set=output_samples)


# Find the simple function approximation
simpleFunP.regular_partition_uniform_distribution_rectangle_scaled(
    data_set=my_discretization, Q_ref=Q_ref, rect_scale=rect_scale,
    cells_per_dimension=1)

# Compute the solution to the stochastic inverse problem
calculateP.prob(my_discretization)

percentile = 1.0
# Sort samples by highest probability density and find how many samples lie in
# the support of the inverse solution.  With the Monte Carlo assumption, this
# also tells us the approximate volume of this support.
(num_samples, _, indices_in_inverse) =\
    postTools.sample_highest_prob(top_percentile=percentile,
                                  sample_set=input_samples, sort=True)

# Print the approximate proportion of the measure of the parameter space defined
# by the support of the inverse density
if comm.rank == 0:
    print('The approximate proportion of the measure of the parameter space defined')
    print('by the support of the inverse density associated with the choice of QoI map is')
    print(np.sum(input_samples.get_volumes()[indices_in_inverse]),
          ' with ', num_samples, ' samples.')



# Copyright (C) 2014-2020 The BET Development Team

# -*- coding: utf-8 -*-
import numpy as np
from dolfin import *
from meshDS import meshDS
from projectKL import projectKL
from poissonRandField import solvePoissonRandomField
import scipy.io as sio


def my_model(parameter_samples):

    # We proceed by loading the mesh and defining the function space for which
    # the eigenfunctions from Compute_Save_KL were computed.

    # Step 1: Set up the Mesh and Function Space
    mesh = Mesh("Lshaped.xml")

    # initialize the mesh to generate connectivity
    mesh.init()

    # Random field is projected on the space of Hat functions in the mesh
    V = FunctionSpace(mesh, "Lagrange", 1)

    # Load the KL expansion information
    KL_mdat = sio.loadmat("KL_expansion")
    KL_eigen_funcs = KL_mdat['KL_eigen_funcs']
    KL_eigen_vals = KL_mdat['KL_eigen_vals']

    # number of parameter samples
    numSamples = parameter_samples.shape[0]

    # number of KL expansion terms.
    numKL = parameter_samples.shape[1]

    # the samples are the coefficients of the KL expansion typically denoted
    # by xi_k
    xi_k = parameter_samples

    '''
    ++++++++++++++++ Steps in Solving Poisson with the KL fields ++++++++++++
    First set up the necessary variables and boundary conditions for the
    problem.

    Then create the QoI maps defined by average values over some part of the
    physical domain.

    Loop through the sample fields and create the permeability defined by the
    exponential of the KL field (i.e., the KL expansion represents the log of
    the permeability).

    For each sample field, call the PoissonRandomField solver.
    +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    '''
    # permeability
    perm_k = Function(V)

    # Create Boundary Conditions -- Dirichlet on left and bottom boundary.
    # Left Dirichlet Bc
    def left_boundary(x, on_boundary):
        """TODO: Docstring for left_boundary.

        :x: TODO
        :on_boundary: TODO
        :returns: TODO

        """
        tol = 1e-14
        return on_boundary and abs(x[0]) < tol
    Gamma_0 = DirichletBC(V, Constant(0.0), left_boundary)

    def bottom_boundary(x, on_boundary):
        """TODO: Docstring for left_boundary.

        :x: TODO
        :on_boundary: TODO
        :returns: TODO

        """
        tol = 1e-14
        return on_boundary and abs(x[1]) < tol
    Gamma_1 = DirichletBC(V, Constant(0.0), bottom_boundary)
    bcs = [Gamma_0, Gamma_1]

    # Setup the QoI class
    class CharFunc(Expression):
        def __init__(self, **kwargs):
            region = kwargs["region"]
            self.a = region[0]
            self.b = region[1]
            self.c = region[2]
            self.d = region[3]

        def eval(self, v, x):
            v[0] = 0
            if (x[0] >= self.a) and (x[0] <= self.b) and (x[1] >= self.c)\
                    and (x[1] <= self.d):
                v[0] = 1
            return v

    # Define the QoI maps
    Chi_1 = CharFunc(degree=1, region=[0.75, 1.25, 7.75, 8.25])
    Chi_2 = CharFunc(degree=1, region=[7.75, 8.25, 0.75, 1.25])

    QoI_samples = np.zeros([numSamples, 2])

    # For each sample solve the PDE
    f = Constant(-1.0)  # forcing of Poisson

    for i in range(0, numSamples):

        print("Sample point number: %g" % i)

        # create a temp array to store logPerm as sum of KL expansions
        # logPerm is log permeability
        logPerm = np.zeros((mesh.num_vertices()), dtype=float)
        for kl in range(0, numKL):
            logPerm += xi_k[i, kl] * \
                np.sqrt(KL_eigen_vals[0, kl]) * KL_eigen_funcs[kl, :]

        # permiability is the exponential of log permeability logPerm
        perm_k_array = 0.1 + np.exp(logPerm)

        # map permiability values to the function space object perm_k
        perm_k.vector()[:] = perm_k_array

        # solve Poisson with this random field using FEM
        u = solvePoissonRandomField(perm_k, V, f, bcs)

        # Compute QoI
        QoI_samples[i, 0] = assemble(u * Chi_1 * dx)
        QoI_samples[i, 1] = assemble(u * Chi_2 * dx)

    return QoI_samples



# Copyright (C) 2014-2020 The BET Development Team

"""
This module contains tests for :module:`bet.postProcess.plotDomains`.


Tests for the execution of plotting parameter and data domains.
"""

import unittest
import os
import glob
import bet
import bet.postProcess.plotDomains as plotDomains
import bet.util as util
import matplotlib.tri as tri
from matplotlib.lines import Line2D
import numpy as np
import numpy.testing as nptest
from bet.Comm import comm
import bet.sample as sample

# local_path = os.path.join(os.path.dirname(bet.__file__),
#        "../test/test_sampling")

local_path = '.'


@unittest.skipIf(comm.size > 1, 'Only run in serial')
class test_plotDomains(unittest.TestCase):
    """
    Test :meth:`bet.postProcess.plotP.calculate_1D_marginal_probs` and
    :meth:`bet.postProcess.plotP.calculate_2D_marginal_probs` for a 2D
    parameter space.
    """

    def setUp(self):
        """
        Set up problem.
        """
        # Create sample_set object for input_samples
        input_samples = sample.sample_set(4)

        input_samples.set_domain(np.array([[0.0, 1.0], [0.0, 1.0],
                                           [0.0, 1.0], [0.0, 1.0]]))
        input_samples.set_values(util.meshgrid_ndim(
            (np.linspace(input_samples.get_domain()[0, 0],
                         input_samples.get_domain()[0, 1], 3),
             np.linspace(input_samples.get_domain()[1, 0],
                         input_samples.get_domain()[1, 1], 3),
             np.linspace(input_samples.get_domain()[2, 0],
                         input_samples.get_domain()[2, 1], 3),
             np.linspace(input_samples.get_domain()[3, 0],
                         input_samples.get_domain()[3, 1], 3))))
        input_samples.set_probabilities(
            (1.0 / float(input_samples.get_values().shape[0]))
            * np.ones((input_samples.get_values().shape[0],)))

        # Check that probabilities and values arrays have same number of
        # entries
        input_samples.check_num()

        # Create sample_set object for output_samples
        output_samples = sample.sample_set(4)
        output_samples.set_values(input_samples.get_values() * 3.0)
        output_samples.set_domain(3.0 * input_samples.get_domain())

        self.disc = sample.discretization(input_samples, output_samples)

        self.filename = "testfigure"

        output_ref_datum = np.mean(output_samples.get_domain(), axis=1)

        bin_size = 0.15 * (np.max(output_samples.get_domain(), axis=1) -
                           np.min(output_samples.get_domain(), axis=1))
        maximum = 1 / np.product(bin_size)

        def ifun(outputs):
            """
            Indicator function.
            :param outputs: outputs
            :type outputs: :class:`numpy.ndarray` of shape (N, ndim)
            :rtype: :class:`numpy.ndarray` of shape (N,)
            :returns: 0 if outside of set or positive number if inside set
            """
            left = np.repeat([output_ref_datum - .5 * bin_size],
                             outputs.shape[0], 0)
            right = np.repeat([output_ref_datum + .5 * bin_size],
                              outputs.shape[0], 0)
            left = np.all(np.greater_equal(outputs, left), axis=1)
            right = np.all(np.less_equal(outputs, right), axis=1)
            inside = np.logical_and(left, right)
            max_values = np.repeat(maximum, outputs.shape[0], 0)
            return inside.astype('float64') * max_values

        self.rho_D = ifun
        self.lnums = [1, 2, 3]
        self.markers = []

        for m in Line2D.markers:
            try:
                if len(m) == 1 and m != ' ':
                    self.markers.append(m)
            except TypeError:
                pass

        self.colors = ('b', 'g', 'r', 'c', 'm', 'y', 'k')

    def tearDown(self):
        """
        Tear Down problem
        """
        # remove any files the we create
        filenames = glob.glob(self.filename + ".*")
        filenames.extend(glob.glob('param_samples_*cs.*'))
        filenames.extend(glob.glob('data_samples_*cs.*'))

        filenames.extend(glob.glob(self.filename + ".*"))
        filenames.extend(glob.glob('param_samples_*cs.*'))
        filenames.extend(glob.glob(os.path.join(local_path,
                                                'data_samples_*cs.*')))

        filenames.append('domain_q1_q2_cs.*')
        filenames.append('domain_q1_q1_cs.*')
        filenames.append('q1_q2_domain_Q_cs.*')
        filenames.append('q1_q1_domain_Q_cs.*')
        figfiles = glob.glob('figs/*')
        figfiles.extend(glob.glob(os.path.join(local_path, 'figs/*')))
        figfiles.extend(glob.glob(os.path.join(local_path, '*.png')))
        figfiles.extend(glob.glob('*.png'))
        filenames.extend(figfiles)

        for f in filenames:
            if os.path.exists(os.path.join(local_path, f)):
                os.remove(os.path.join(local_path, f))
            if os.path.exists(f):
                os.remove(f)
        if os.path.exists("figs"):
            os.rmdir("figs")

    def test_scatter_2D(self):
        """
        Test :meth:`bet.postProcess.plotDomains.scatter_2D`
        """
        sample_nos = [None, 25]
        p_ref = [None, self.disc._input_sample_set.get_values()[4, [0, 1]]]
        #p_ref = [None, self.samples[4, [0, 1]]]
        for sn, pr in zip(sample_nos, p_ref):
            self.check_scatter_2D(sn, pr, True)

    def check_scatter_2D(self, sample_nos, p_ref, save):
        """
        Check to see that the :meth:`bet.postTools.plotDomains.scatter_2D` ran
        without generating an error.
        """
        try:
            input_sample_set_temp = sample.sample_set(2)
            input_sample_set_temp.set_values(
                self.disc._input_sample_set.get_values()[:, [0, 1]])
            plotDomains.scatter_2D(
                input_sample_set_temp,
                sample_nos,
                self.disc._input_sample_set.get_probabilities(),
                p_ref, save, False, 'XLABEL', 'YLABEL', self.filename)
            go = True
        except (RuntimeError, TypeError, NameError):
            go = False

        nptest.assert_equal(go, True)

    def check_scatter_2D_io(self, sample_nos, p_ref, save):
        """
        Check to see that the :meth:`bet.postTools.plotDomains.scatter_2D_input` ran
        without generating an error.
        """
        try:
            input_sample_set_temp = sample.sample_set(2)
            input_sample_set_temp.set_values(
                self.disc._input_sample_set.get_values()[:, [0, 1]])

            disc = sample.discretization(input_sample_set=input_sample_set_temp,
                                         output_sample_set=input_sample_set_temp)

            plotDomains.scatter_2D_input(
                disc,
                sample_nos,
                self.disc._input_sample_set.get_probabilities(),
                p_ref, save, False, 'XLABEL', 'YLABEL', self.filename)
            plotDomains.scatter_2D_output(
                disc,
                sample_nos,
                self.disc._input_sample_set.get_probabilities(),
                p_ref, save, False, 'XLABEL', 'YLABEL', self.filename)
            go = True
        except (RuntimeError, TypeError, NameError):
            go = False

        nptest.assert_equal(go, True)

    def test_scatter_3D(self):
        """
        Test :meth:`bet.postProcess.plotDomains.scatter_3D`
        """
        sample_nos = [None, [25]]
        p_ref = [None, self.disc._input_sample_set.get_values()[4, :]]
        for sn, pr in zip(sample_nos, p_ref):
            self.check_scatter_3D(sn, pr, True)

    def check_scatter_3D(self, sample_nos, p_ref, save):
        """
        Check to see that the :meth:`bet.postTools.plotDomains.scatter_3D` ran
        without generating an error.
        """
        try:
            input_sample_set_temp = sample.sample_set(3)
            input_sample_set_temp.set_values(
                self.disc._input_sample_set.get_values()[:, [0, 1, 2]])
            plotDomains.scatter_3D(
                input_sample_set_temp,
                sample_nos,
                self.disc._input_sample_set.get_probabilities(),
                p_ref, save, False, 'XLABEL', 'YLABEL', 'ZLABEL', self.filename)
            go = True
        except (RuntimeError, TypeError, NameError):
            go = False

        nptest.assert_equal(go, True)

    def check_scatter_3D_io(self, sample_nos, p_ref, save):
        """
        Check to see that the :meth:`bet.postTools.plotDomains.scatter_3D_input` ran
        without generating an error.
        """
        try:
            input_sample_set_temp = sample.sample_set(3)
            input_sample_set_temp.set_values(
                self.disc._input_sample_set.get_values()[:, [0, 1, 2]])
            disc = sample.discretization(input_sample_set=input_sample_set_temp,
                                         output_sample_set=input_sample_set_temp)
            plotDomains.scatter_3D_input(
                disc,
                sample_nos,
                self.disc._input_sample_set.get_probabilities(),
                p_ref, save, False, 'XLABEL', 'YLABEL', 'ZLABEL', self.filename)

            plotDomains.scatter_3D_output(
                disc,
                sample_nos,
                self.disc._input_sample_set.get_probabilities(),
                p_ref, save, False, 'XLABEL', 'YLABEL', 'ZLABEL', self.filename)
            go = True
        except (RuntimeError, TypeError, NameError):
            go = False

        nptest.assert_equal(go, True)

    def test_show_param(self):
        """
        Test :meth:`bet.postProcess.plotDomains.scatter_rhoD`
        """
        sample_nos = [None, [25]]
        samples = [self.disc._input_sample_set.get_values(),
                   self.disc._input_sample_set.get_values()[:, [0, 1]],
                   self.disc._input_sample_set.get_values()[:, [0, 1, 2]]]
        lnums = [None, self.lnums]

        for sample in samples:
            showdim = [None]
            if sample.shape[0] > 2:
                showdim.append(2)
            if sample.shape[0] > 3:
                showdim.append(3)
            for sd in showdim:
                p_ref = [None, sample[4, :]]
                for ln, sn, pr in zip(lnums, sample_nos, p_ref):
                    self.check_show_param(sample, sn, pr, True, ln, sd)

    def check_show_param(self, samples, sample_nos, p_ref, save, lnums,
                         showdim):
        """
        Check to see that the :meth:`bet.postTools.plotDomains.scatter_rhoD` ran
        without generating an error.
        """
        try:
            input_sample_set_temp = sample.sample_set(samples.shape[1])
            input_sample_set_temp.set_values(samples)
            disc_obj_temp = sample.discretization(input_sample_set_temp,
                                                  self.disc._output_sample_set)
            plotDomains.scatter_rhoD(disc_obj_temp, p_ref, sample_nos, 'input',
                                     self.rho_D, lnums, None, showdim, save, False)
            go = True
        except (RuntimeError, TypeError, NameError) as error:
            print("ERROR:", error)
            print("samples shape:", samples.shape)
            print("param ref:", p_ref)
            print("samples nums:", sample_nos)
            print("save:", save)
            print("lnums:", lnums)
            print("showdim:", showdim)
            go = False

        nptest.assert_equal(go, True)

    def test_show_data(self):
        """
        Test :meth:`bet.postProcess.plotDomains.scatter_rhoD`
        """
        sample_nos = [None, 25]
        data_sets = [self.disc._output_sample_set.get_values(),
                     self.disc._output_sample_set.get_values()[:, [0, 1]]]
        qnums = [None, [0, 1, 2]]  # self.lnums]

        for data, qn, sn in zip(data_sets, qnums, sample_nos):
            showdim = [None]
            if data.shape[0] > 2:
                showdim.append(2)
            if data.shape[0] > 3:
                showdim.append(3)
            Q_ref = [None, data[4, :]]
            for sd, qr in zip(showdim, Q_ref):
                self.check_show_data(data, sn, qr, True, qn, sd)

    def check_show_data(self, data, sample_nos, q_ref, save, qnums, showdim):
        """
        Check to see that the :meth:`bet.postTools.plotDomains.scatter_rhoD` ran
        without generating an error.
        """
        try:
            if data.shape[1] == 4:
                data_obj_temp = sample.sample_set(4)
                data_obj_temp.set_values(data)
                plotDomains.scatter_rhoD(data_obj_temp, q_ref, sample_nos,
                                         'output', self.rho_D, qnums, None, showdim, save,
                                         False)
            else:
                data_obj_temp = sample.sample_set(data.shape[1])
                data_obj_temp.set_values(data)
                plotDomains.scatter_rhoD(data_obj_temp, q_ref, sample_nos,
                                         None, None, qnums, None, showdim, save, False)
            go = True
        except (RuntimeError, TypeError, NameError):
            print("ERROR")
            print("data shape:", data.shape)
            print("data ref:", q_ref)
            print("samples nums:", sample_nos)
            print("save:", save)
            print("qnums:", qnums)
            print("showdim:", showdim)
            go = False
        nptest.assert_equal(go, True)

    def test_show_data_domain_2D(self):
        """
        Test :meth:`bet.postProces.plotDomains.show_data_domain_2D`
        """
        ref_markers = [None, self.markers]
        ref_colors = [None, self.colors]
        filenames = [None, ['domain_q1_q1_cs', 'q1_q2_domain_Q_cs']]

        for rm, rc, fn in zip(ref_markers, ref_colors, filenames):
            self.check_show_data_domain_2D(rm, rc, None, True, fn)

    def check_show_data_domain_2D(self, ref_markers, ref_colors, triangles,
                                  save, filenames):
        """
        Check to see that the
        :meth:`bet.postTools.plotDomains.show_data_domain_2D` ran
        without generating an error.
        """
        Q_ref = self.disc._output_sample_set.get_values()[:, [0, 1]]
        Q_ref = Q_ref[[1, 4], :]

        data_obj_temp = sample.sample_set(2)
        data_obj_temp.set_values(
            self.disc._output_sample_set.get_values()[:, [0, 1]])
        disc_obj_temp = sample.discretization(
            self.disc._input_sample_set, data_obj_temp)

        try:
            plotDomains.show_data_domain_2D(
                disc_obj_temp, Q_ref,
                ref_markers, ref_colors, save=save,
                filenames=filenames)
            go = True
        except (RuntimeError, TypeError, NameError):
            go = False

        nptest.assert_equal(go, True)

    def test_show_data_domain_multi(self):
        """
        Test :meth:`bet.postProcess.plotDomains.show_data_domain_multi`
        """
        if not os.path.exists('figs/'):
            os.mkdir('figs/')

        Q_nums = [None, [1, 2], [1, 2, 3]]
        ref_markers = [None, self.markers]
        ref_colors = [None, self.colors]

        for rm, rc in zip(ref_markers, ref_colors):
            for qn in Q_nums:
                showdim = [None, 1]
                if qn and len(qn) > 2:
                    showdim.extend(['all', 'ALL'])
                for sd in showdim:
                    self.check_show_data_domain_multi(rm, rc, qn, sd)

    def check_show_data_domain_multi(self, ref_markers, ref_colors, Q_nums,
                                     showdim):
        """
        Check to see that the
        :meth:`bet.postTools.plotDomains.show_data_domain_multi` ran
        without generating an error.
        """
        Q_ref = self.disc._output_sample_set.get_values()[[4, 2], :]
        try:
            plotDomains.show_data_domain_multi(
                self.disc,
                Q_ref, Q_nums, ref_markers=ref_markers,
                ref_colors=ref_colors, showdim=showdim)
            go = True
        except (RuntimeError, TypeError, NameError):
            go = False
        nptest.assert_equal(go, True)

    def test_scatter_2D_multi(self):
        """
        Test :meth:`bet.postTools.plotDomins.scatter_2D_multi`
        """
        if not os.path.exists('figs/'):
            os.mkdir('figs/')
        try:
            input_sample_set_temp = sample.sample_set(3)
            input_sample_set_temp.set_values(
                self.disc._input_sample_set.get_values()[:, [0, 1, 2]])

            plotDomains.scatter_2D_multi(input_sample_set_temp)
            go = True
        except (RuntimeError, TypeError, NameError):
            go = False

        nptest.assert_equal(go, True)



# Copyright (C) 2014-2020 The BET Development Team

"""
This module provides methods for generating and using surrogate models.
"""
import numpy as np
import bet.sample as sample
import bet.calculateP.calculateError as calculateError
import bet.calculateP.calculateP as calculateP
from bet.Comm import comm, MPI


class piecewise_polynomial_surrogate(object):
    """
    This class provides methods for generating a piecewise polynomial
    surrogate.
    """

    def __init__(self, input_disc):
        """
        Initializes a piecewise polynomial surrogate based on
        existing input discretization.

        :param discretization: An object containing the discretization
            information.
        :type discretization: :class:`bet.sample.discretization`

        """
        if not isinstance(input_disc, sample.discretization):
            msg = "The argument must be of type bet.sample.discretization."
            raise calculateError.wrong_argument_type(msg)
        input_disc.check_nums()
        self.input_disc = input_disc
        self.input_disc._input_sample_set.local_to_global()
        self.input_disc._output_sample_set.local_to_global()

    def generate_for_input_set(self, input_sample_set, order=0):
        """
        Generates a surrogate discretization based on the input discretization,
        for a user-defined input sample set. The output sample set values
        and error estimates are piecewise polynomially defined over input sample
        set cells from the input discretization. For order 0, both are piecewise
        constant. For order 1, values are piecewise linear (assuming Jacobians
        exist), and error estimates are piecewise constant.

        :param input_sample_set: input sample set for surrogate discretization
        :type set_old: :class:`~bet.sample.sample_set_base`
        :param order: Polynomial order
        :type order: int

        :rtype: :class:`~bet.sample.discretization`
        :returns: discretization defining the surrogate model

        """
        # Check inputs
        if order not in [0, 1]:
            msg = "Order must be 0 or 1."
            raise calculateError.wrong_argument_type(msg)
        input_sample_set.check_num()
        if input_sample_set._dim != self.input_disc._input_sample_set._dim:
            msg = "Dimensions of input sets are not equal."
            raise sample.dim_not_matching(msg)

        # Give properties from input discretization.
        if input_sample_set._domain is None:
            if self.input_disc._input_sample_set._domain is not None:
                input_sample_set.set_domain(self.input_disc.
                                            _input_sample_set._domain)
        if input_sample_set._p_norm is None:
            if self.input_disc._input_sample_set._p_norm is not None:
                input_sample_set.set_p_norm(self.input_disc.
                                            _input_sample_set._p_norm)

        # Setup dummy discretizion to get pointers
        # Assumes Voronoi sample set for now
        output_sample_set = sample.sample_set(self.input_disc.
                                              _output_sample_set._dim)
        self.dummy_disc = self.input_disc.copy()
        self.dummy_disc.set_emulated_input_sample_set(input_sample_set)
        self.dummy_disc.set_emulated_ii_ptr(globalize=False)

        if order == 0:
            # define new values based on piecewise constants
            new_values_local = self.input_disc._output_sample_set.\
                _values[self.dummy_disc._emulated_ii_ptr_local]
            output_sample_set.set_values_local(new_values_local)
        elif order == 1:
            # define new values based on piecewise linears using Jacobians
            if self.input_disc._input_sample_set._jacobians is None:
                if self.input_disc._input_sample_set._jacobians_local is None:
                    msg = "The input discretization must"
                    msg += " have jacobians defined."
                    raise calculateError.wrong_argument_type(msg)
                else:
                    self.input_disc._input_sample_set.local_to_global()

            jac_local = self.input_disc._input_sample_set._jacobians[
                self.dummy_disc._emulated_ii_ptr_local]
            diff_local = self.input_disc._input_sample_set._values[
                self.dummy_disc._emulated_ii_ptr_local] - \
                input_sample_set._values_local
            new_values_local = self.input_disc._output_sample_set._values[
                self.dummy_disc._emulated_ii_ptr_local]
            new_values_local += np.einsum('ijk,ik->ij', jac_local, diff_local)
            output_sample_set.set_values_local(new_values_local)

        # if they exist, define error estimates with piecewise constants
        if self.input_disc._output_sample_set._error_estimates is not None:
            new_ee = self.input_disc._output_sample_set._error_estimates[
                self.dummy_disc._emulated_ii_ptr_local]
            output_sample_set.set_error_estimates_local(new_ee)
        # create discretization object for the surrogate
        self.surrogate_discretization = sample.discretization(input_sample_set=input_sample_set, output_sample_set=output_sample_set,
                                                              output_probability_set=self.input_disc._output_probability_set)
        return self.surrogate_discretization

    def calculate_prob_for_sample_set_region(self, s_set,
                                             regions, update_input=True):
        """
        Solves stochastic inverse problem based on surrogate points and the
        MC assumption. Calculates the probability of a regions of input space
        and error estimates for those probabilities.

        :param: s_set: sample set for which to calculate error
        :type s_set: :class:`bet.sample.sample_set_base`
        :param region: list of regions of s_set for which to calculate error
        :type region: list
        :param update_input: whether or not to update probabilities and
            errror identifiers for input discretization
        :type update_input: bool

        :rtype: tuple
        :returns: (probabilities, ``error_estimates``), the probability and
            error estimates for the region

        """
        if not hasattr(self, 'surrogate_discretization'):
            msg = "surrogate discretization has not been created"
            raise calculateError.wrong_argument_type(msg)
        if not isinstance(s_set, sample.sample_set_base):
            msg = "s_set must be of type bet.sample.sample_set_base"
            raise calculateError.wrong_argument_type(msg)

        # Calculate probability of region
        if self.surrogate_discretization._input_sample_set._volumes_local\
                is None:
            self.surrogate_discretization._input_sample_set.\
                estimate_volume_mc(globalize=False)
        calculateP.prob(self.surrogate_discretization, globalize=False)
        prob_new_values = calculateP.prob_from_sample_set(
            self.surrogate_discretization._input_sample_set, s_set)

        # Calculate for each region
        probabilities = []
        error_estimates = []
        for region in regions:
            marker = np.equal(s_set._region, region)
            probability = np.sum(prob_new_values[marker])

            # Calculate error estimate for region
            model_error = calculateError.model_error(
                self.surrogate_discretization)
            error_estimate = model_error.calculate_for_sample_set_region_mc(
                s_set, region)
            probabilities.append(probability)
            error_estimates.append(error_estimate)
        # Update input only if 1 region is given
        if update_input:
            num = self.input_disc._input_sample_set.check_num()
            prob = np.zeros((num,))
            error_id = np.zeros((num,))
            for i in range(num):
                Itemp = np.equal(self.dummy_disc._emulated_ii_ptr_local, i)
                prob_sum = np.sum(self.surrogate_discretization.
                                  _input_sample_set._probabilities_local[Itemp])
                prob[i] = comm.allreduce(prob_sum, op=MPI.SUM)
                error_id_sum = np.sum(self.surrogate_discretization.
                                      _input_sample_set._error_id_local[Itemp])
                error_id[i] = comm.allreduce(error_id_sum, op=MPI.SUM)
            self.input_disc._input_sample_set.set_probabilities(prob)
            self.input_disc._input_sample_set.set_error_id(error_id)

        return (probabilities, error_estimates)



# Copyright (C) 2014-2020 The BET Development Team

"""
This module contains functions for choosing optimal sets of QoIs to use in the
stochastic inverse problem.
"""
import logging
from itertools import combinations
import numpy as np
from scipy import stats
from bet.Comm import comm
import bet.sample
import bet.util as util


def calculate_avg_measure(input_set, qoi_set=None, bin_measure=None):
    r"""
    If you are using ``bin_ratio`` to define the hyperrectangle in the output
    space you must give this method gradient vectors normalized with
    respect to the 1-norm.  If you are using ``bin_size`` to define the
    hyperrectangle in the output space you must give this method the original
    gradient vectors. If you also give a ``bin_measure``, this method will
    approximate the measure of the region of non-zero probability in the inverse
    solution.
    Given gradient vectors at some points (centers) in the input space and
    given a specific set of QoIs, calculate the expected measure of the
    inverse image of a box in the data space using local linear approximations
    of the map Q.

    :param input_set: The input sample set.  Make sure the attribute _jacobians
        is not None
    :type input_set: :class:`~bet.sample.sample_set`
    :param list qoi_set: List of QoI indices
    :param float bin_measure: The measure of the output_dim hyperrectangle to
        invert into the input space

    :rtype: tuple
    :returns: (avg_measure, singvals) where avg_measure is a float and singvals
        has shape (num_centers, output_dim)

    """

    if input_set._jacobians is None:
        raise ValueError("You must have jacobians to use this method.")
    if qoi_set is None:
        G = input_set._jacobians
    else:
        G = input_set._jacobians[:, qoi_set, :]
    if G.shape[1] > G.shape[2]:
        raise ValueError("Measure is not defined for more outputs than inputs.\
            Try adding a qoi_set to evaluate the measure of.")

    # If no measure is given, we consider how this set of QoIs will change the
    # measure of the unit hypercube.
    if bin_measure is None:
        bin_measure = 1.0

    # Calculate the singular values of the matrix formed by the gradient
    # vectors of each QoI map.  This gives a set of singular values for each
    # center.
    singvals = np.linalg.svd(G, compute_uv=False)

    # Find the average product of the singular values over each center, then use
    # this to compute the average measure of the inverse solution.
    avg_prod_singvals = np.mean(np.prod(singvals, axis=1))
    if avg_prod_singvals == 0:
        avg_measure = np.inf
    else:
        avg_measure = bin_measure / avg_prod_singvals

    return avg_measure, singvals


def calculate_avg_skewness(input_set, qoi_set=None):
    r"""
    Given gradient vectors at some points (centers) in the input space and
    given a specific set of QoIs, caculate the average skewness of the arrays
    formed by the gradient vectors of each QoI map at each center.

    :param input_set: The input sample set.  Make sure the attribute _jacobians
        is not None
    :type input_set: :class:`~bet.sample.sample_set`
    :param list qoi_set: List of QoI indices
    :rtype: tuple
    :returns: (hmean_skewG, skewgi) where hmean_skewG is the harmonic mean of
        skewness at each center in the input space (float) and skewgi
        has shape (num_centers, output_dim)
    """

    if input_set._jacobians is None:
        raise ValueError("You must have jacobians to use this method.")
    if qoi_set is None:
        G = input_set._jacobians
    else:
        G = input_set._jacobians[:, qoi_set, :]
    if G.shape[1] > G.shape[2]:
        msg = "Skewness is not defined for more outputs than inputs."
        msg += " Try adding a qoi_set to evaluate the skewness of."
        raise ValueError(msg)

    num_centers = G.shape[0]
    output_dim = G.shape[1]

    # Calculate the singular values of the matrix formed by the gradient
    # vectors of each QoI map.  This gives a set of singular values for each
    # center.
    singvals = np.linalg.svd(G, compute_uv=False)

    # The measure of the parallelepipeds defined by the rows of each Jacobian
    muG = np.tile(np.prod(singvals, axis=1), [output_dim, 1]).transpose()

    # Calculate the measure of the parallelepipeds defined by the rows of each
    # Jacobian if we remove the i'th row.
    muGi = np.zeros([num_centers, output_dim])
    for i in range(G.shape[1]):
        muGi[:, i] = np.prod(np.linalg.svd(np.delete(G, i, axis=1),
                                           compute_uv=False), axis=1)

    # Find the norm of each gradient vector
    normgi = np.linalg.norm(G, axis=2)

    # Find the norm of the new vector, giperp, that is perpendicular to the span
    # of the other vectors and defines a parallelepiped of the same measure.
    normgiperp = muG / muGi

    # We now calculate the local skewness
    skewgi = np.zeros([num_centers, output_dim])

    # The local skewness is calculated for nonzero giperp
    skewgi[normgiperp != 0] = normgi[normgiperp != 0] / \
        normgiperp[normgiperp != 0]

    # If giperp is the zero vector, it is not GD from the rest of the gradient
    # vectors, so the skewness is infinity.
    skewgi[normgiperp == 0] = np.inf

    # If the norm of giperp is infinity, then the rest of the vector were not GD
    # to begin with, so skewness is infinity.
    skewgi[normgiperp == np.inf] = np.inf

    # The local skewness is the max skewness of each vector relative the rest
    skewG = np.max(skewgi, axis=1)
    skewG[np.isnan(skewG)] = np.inf

    # We may have values equal to infinity, so we consider the harmonic mean.
    hmean_skewG = stats.hmean(skewG)

    return hmean_skewG, skewgi


def calculate_avg_condnum(input_set, qoi_set=None):
    r"""
    Given gradient vectors at some points (centers) in the input space and
    given a specific set of QoIs, caculate the average condition number of the
    matrices formed by the gradient vectors of each QoI map at each center.

    :param input_set: The input sample set.  Make sure the attribute _jacobians
        is not None
    :type input_set: :class:`~bet.sample.sample_set`
    :param list qoi_set: List of QoI indices

    :rtype: tuple
    :returns: (condnum, singvals) where condnum is a float and singvals
        has shape (num_centers, output_dim)

    """

    if input_set._jacobians is None:
        raise ValueError("You must have jacobians to use this method.")
    if qoi_set is None:
        G = input_set._jacobians
    else:
        G = input_set._jacobians[:, qoi_set, :]
    if G.shape[1] > G.shape[2]:
        msg = "Condition number is not defined for more outputs than inputs."
        msg += " Try adding a qoi_set to evaluate the condition number of."
        raise ValueError(msg)

    # Calculate the singular values of the matrix formed by the gradient
    # vectors of each QoI map.  This gives a set of singular values for each
    # center.
    singvals = np.linalg.svd(G, compute_uv=False)
    indz = singvals[:, -1] == 0
    if np.sum(indz) == singvals.shape[0]:
        hmean_condnum = np.inf
    else:
        singvals[indz, 0] = np.inf
        singvals[indz, -1] = 1
        condnums = singvals[:, 0] / singvals[:, -1]
        hmean_condnum = stats.hmean(condnums)

    return hmean_condnum, singvals


def chooseOptQoIs(input_set, qoiIndices=None, num_qois_return=None,
                  num_optsets_return=None, inner_prod_tol=1.0, measure=False,
                  remove_zeros=True):
    r"""
    Given gradient vectors at some points (centers) in the parameter space, a
    set of QoIs to choose from, and the number of desired QoIs to return, this
    method returns the ``num_optsets_return`` best sets of QoIs with with
    repsect to either the average measure of the matrix formed by the
    gradient vectors of each QoI map, OR the average skewness of the inverse
    image of this set of QoIs, computed as the product of the singular values
    of the same matrix.  This method is brute force, i.e., if the method is
    given 10,000 QoIs and told to return the N best sets of 3, it will check all
    10,000 choose 3 possible sets.  See chooseOptQoIs_large for a less
    computationally expensive approach.

    :param input_set: The input sample set.  Make sure the attribute _jacobians
        is not None
    :type input_set: :class:`~bet.sample.sample_set`
    :param qoiIndices: Set of QoIs to consider.  Default is
        xrange(0, input_set._jacobians.shape[1])
    :type qoiIndices: :class:`np.ndarray` of size (1, num QoIs to consider)
    :param int num_qois_return: Number of desired QoIs to use in the
        inverse problem.  Default is input_dim
    :param int num_optsets_return: Number of best sets to return
        Default is 10
    :param boolean measure: If measure is True, use ``calculate_avg_measure``
        to determine optimal QoIs, else use ``calculate_avg_skewness``
    :param boolean remove_zeros: If True, ``find_unique_vecs`` will remove any
        QoIs that have a zero gradient

    :rtype: `np.ndarray` of shape (num_optsets_returned, num_qois_returned + 1)
    :returns: measure_skewness_indices_mat

    """

    (measure_skewness_indices_mat, _) = chooseOptQoIs_verbose(input_set,
                                                              qoiIndices, num_qois_return, num_optsets_return, inner_prod_tol,
                                                              measure, remove_zeros)

    return measure_skewness_indices_mat


def chooseOptQoIs_verbose(input_set, qoiIndices=None, num_qois_return=None,
                          num_optsets_return=None, inner_prod_tol=1.0, measure=False,
                          remove_zeros=True):
    r"""
    Given gradient vectors at some points (centers) in the parameter space, a
    set of QoIs to choose from, and the number of desired QoIs to return, this
    method returns the ``num_optsets_return`` best sets of QoIs with with
    repsect to either the average measure of the matrix formed by the
    gradient vectors of each QoI map, OR the average skewness of the inverse
    image of this set of QoIs, computed as the product of the singular values
    of the same matrix.  This method is brute force, i.e., if the method is
    given 10,000 QoIs and told to return the N best sets of 3, it will check all
    10,000 choose 3 possible sets.  See chooseOptQoIs_large for a less
    computationally expensive approach.

    :param input_set: The input sample set.  Make sure the attribute _jacobians
        is not None
    :type input_set: :class:`~bet.sample.sample_set`
    :param qoiIndices: Set of QoIs to consider.  Default is
        xrange(0, input_set._jacobians.shape[1])
    :type qoiIndices: :class:`np.ndarray` of size (1, num QoIs to consider)
    :param int num_qois_return: Number of desired QoIs to use in the
        inverse problem.  Default is input_dim
    :param int num_optsets_return: Number of best sets to return
        Default is 10
    :param boolean measure: If measure is True, use ``calculate_avg_measure``
        to determine optimal QoIs, else use ``calculate_avg_skewness``
    :param boolean remove_zeros: If True, ``find_unique_vecs`` will remove any
        QoIs that have a zero gradient

    :rtype: `np.ndarray` of shape (num_optsets_returned, num_qois_returned + 1)
    :returns: measure_skewness_indices_mat

    """

    G = input_set._jacobians
    if G is None:
        raise ValueError("You must have jacobians to use this method.")
    input_dim = input_set._dim
    num_centers = G.shape[0]

    if qoiIndices is None:
        qoiIndices = range(0, G.shape[1])
    if num_qois_return is None:
        num_qois_return = input_dim
    if num_optsets_return is None:
        num_optsets_return = 10

    # Remove QoIs that have zero gradients at any of the centers
    qoiIndices = find_unique_vecs(input_set, inner_prod_tol, qoiIndices,
                                  remove_zeros)

    # Find all posible combinations of QoIs
    if comm.rank == 0:
        qoi_combs = np.array(list(combinations(list(qoiIndices),
                                               num_qois_return)))
        logging.info('Possible sets of QoIs : {}'.format(qoi_combs.shape[0]))
        qoi_combs = np.array_split(qoi_combs, comm.size)
    else:
        qoi_combs = None

    # Scatter them throughout the processors
    qoi_combs = comm.scatter(qoi_combs, root=0)

    # For each combination, check the skewness and keep the sets
    # that have the smallest skewness
    measure_skewness_indices_mat = np.zeros([num_optsets_return,
                                             num_qois_return + 1])
    measure_skewness_indices_mat[:, 0] = np.inf
    optsingvals_tensor = np.zeros([num_centers, num_qois_return,
                                   num_optsets_return])
    for qoi_set in range(len(qoi_combs)):
        if measure == False:
            (current_measskew, singvals) = calculate_avg_skewness(input_set,
                                                                  qoi_combs[qoi_set])
        else:
            (current_measskew, singvals) = calculate_avg_measure(input_set,
                                                                 qoi_combs[qoi_set])

        if current_measskew < measure_skewness_indices_mat[-1, 0]:
            measure_skewness_indices_mat[-1, :] = np.append(np.array(
                [current_measskew]), qoi_combs[qoi_set])
            order = measure_skewness_indices_mat[:, 0].argsort()
            measure_skewness_indices_mat = measure_skewness_indices_mat[order]

            optsingvals_tensor[:, :, -1] = singvals
            optsingvals_tensor = optsingvals_tensor[:, :, order]

    # Wait for all processes to get to this point
    comm.Barrier()

    # Gather the best sets and skewness values from each processor
    measure_skewness_indices_mat = np.array(comm.gather(
        measure_skewness_indices_mat, root=0))
    optsingvals_tensor = np.array(comm.gather(optsingvals_tensor, root=0))

    # Find the num_optsets_return smallest skewness values from all processors
    if comm.rank == 0:
        measure_skewness_indices_mat = measure_skewness_indices_mat.reshape(
            num_optsets_return * comm.size, num_qois_return + 1)
        optsingvals_tensor = optsingvals_tensor.reshape(num_centers,
                                                        num_qois_return, num_optsets_return * comm.size)
        order = measure_skewness_indices_mat[:, 0].argsort()

        measure_skewness_indices_mat = measure_skewness_indices_mat[order]
        measure_skewness_indices_mat = measure_skewness_indices_mat[
            :num_optsets_return, :]

        optsingvals_tensor = optsingvals_tensor[:, :, order]
        optsingvals_tensor = optsingvals_tensor[:, :, :num_optsets_return]

    measure_skewness_indices_mat = comm.bcast(measure_skewness_indices_mat,
                                              root=0)
    optsingvals_tensor = comm.bcast(optsingvals_tensor, root=0)

    return (measure_skewness_indices_mat, optsingvals_tensor)


def find_unique_vecs(input_set, inner_prod_tol, qoiIndices=None,
                     remove_zeros=True):
    r"""
    Given gradient vectors at each center in the parameter space, sort throught
    them and remove any QoI that has a zero vector at any center, then remove
    one from any pair of QoIs that have an average inner product greater than
    some tolerance, i.e., an average angle between the two vectors smaller than
    some tolerance.

    :param input_set: The input sample set.  Make sure the attribute _jacobians
        is not None
    :type input_set: :class:`~bet.sample.sample_set`
    :param float inner_prod_tol: Maximum acceptable average inner product
        between two QoI maps
    :param qoiIndices: Set of QoIs to consider
    :type qoiIndices: :class:'`np.ndarray` of size (1, num QoIs to consider)
    :param boolean remove_zeros: If True, ``find_unique_vecs`` will remove any
        QoIs that have a zero gradient vector at atleast one point in
        :math:`\Lambda`

    :rtype: `np.ndarray` of shape (num_unique_vecs, 1)
    :returns: unique_vecs

    """

    if input_set._jacobians is None:
        raise ValueError("You must have jacobians to use this method.")
    input_dim = input_set._dim
    G = input_set._jacobians
    if qoiIndices is None:
        qoiIndices = range(0, G.shape[1])

    # Normalize the gradient vectors with respect to the 2-norm so the inner
    # product tells us about the angle between the two vectors.
    norm_G = np.linalg.norm(G, ord=2, axis=2)

    # Remove any QoI that has a zero vector at atleast one of the centers.
    if remove_zeros:
        indz = np.array([])
        for i in range(norm_G.shape[1]):
            if np.sum(norm_G[:, i] == 0) > 0:
                indz = np.append(indz, i)
    else:
        indz = []

    # If it is a zero vector (has 0 norm), set norm=1, avoid divide by zero
    norm_G[norm_G == 0] = 1.0

    # Normalize each gradient vector
    G = G / np.tile(norm_G, (input_dim, 1, 1)).transpose(1, 2, 0)

    if comm.rank == 0:
        logging.info('*** find_unique_vecs ***')
        logging.info('num_zerovec : {} of ({}) original QoIs'.
                     format(len(indz), G.shape[1]))
        logging.info('Possible QoIs : {}'.format(len(qoiIndices) - len(indz)))
    qoiIndices = list(set(qoiIndices) - set(indz))

    # Find all num_qois choose 2 pairs of QoIs
    qoi_combs = np.array(list(combinations(list(qoiIndices), 2)))

    # For each pair, check the angle between the vectors and throw out the
    # second QoI if the angle is below some tolerance.  At this point all the
    # vectors are normalized, so the inner product will be between -1 and 1.
    repeat_vec = np.array([])
    for qoi_set in range(len(qoi_combs)):
        curr_set = qoi_combs[qoi_set]

        # If neither of the current QoIs are in the repeat_vec, test them
        if curr_set[0] not in repeat_vec and curr_set[1] not in repeat_vec:
            curr_inner_prod = np.sum(G[:, curr_set[0], :] *
                                     G[:, curr_set[1], :]) / G.shape[0]

            # If the innerprod>tol, throw out the second QoI
            if np.abs(curr_inner_prod) > inner_prod_tol:
                repeat_vec = np.append(repeat_vec, qoi_combs[qoi_set, -1])

    unique_vecs = np.array(list(set(qoiIndices) - set(repeat_vec)))
    if comm.rank == 0:
        logging.info('Unique QoIs : {}'.format(unique_vecs.shape[0]))

    return unique_vecs


def find_good_sets(input_set, good_sets_prev, unique_indices,
                   num_optsets_return, measskew_tol, measure):
    r"""

    .. todo::  Use the idea we only know vectors are with 10% accuracy to guide
        inner_prod tol and skewness_tol.

    Given gradient vectors at each center in the parameter space and given
    good sets of size (n - 1), return good sets of size n.  That is, return
    sets of size n that have average measure(skewness) less than some tolerance.

    :param input_set: The input sample set.  Make sure the attribute _jacobians
        is not None.
    :type input_set: :class:`~bet.sample.sample_set`
    :param good_sets_prev: Good sets of QoIs of size n - 1.
    :type good_sets_prev: :class:`np.ndarray` of size (num_good_sets_prev, n -
        1)
    :param unique_indices: Unique QoIs to consider.
    :type unique_indices: :class:`np.ndarray` of size (num_unique_qois, 1)
    :param int num_optsets_return: Number of best sets to return
    :param float measskew_tol: Throw out all sets of QoIs with average
        measure(skewness) number greater than this.
    :param boolean measure: If measure is True, use ``calculate_avg_measure``
        to determine optimal QoIs, else use ``calculate_avg_skewness``

    :rtype: tuple
    :returns: (good_sets, best_sets, optsingvals_tensor) where good sets has
        size (num_good_sets, n), best sets has size (num_optsets_return,
        n + 1) and optsingvals_tensor has size (num_centers, n, input_dim)

    """

    if input_set._jacobians is None:
        raise ValueError("You must have jacobians to use this method.")

    num_centers = input_set._jacobians.shape[0]
    num_qois_return = good_sets_prev.shape[1] + 1
    comm.Barrier()

    # Initialize best sets and set all skewness values large
    best_sets = np.zeros([num_optsets_return, num_qois_return + 1])
    best_sets[:, 0] = np.inf
    good_sets = np.zeros([1, num_qois_return])
    count_qois = 0
    optsingvals_tensor = np.zeros([num_centers, num_qois_return,
                                   num_optsets_return])

    # For each good set of size (n - 1), find the possible sets of size n and
    # compute the average skewness of each
    count_qois = 0
    for i in range(good_sets_prev.shape[0]):
        min_ind = np.max(good_sets_prev[i, :])
        # Find all possible combinations of QoIs that include this set of
        # (n - 1)
        if comm.rank == 0:
            inds_notin_set = util.fix_dimensions_vector_2darray(list(set(
                unique_indices) - set(good_sets_prev[i, :])))

            # Choose only the QoI indices > min_ind so we do not repeat sets
            inds_notin_set = util.fix_dimensions_vector_2darray(inds_notin_set[
                inds_notin_set > min_ind])
            qoi_combs = util.fix_dimensions_vector_2darray(np.append(np.tile(
                good_sets_prev[i, :], [inds_notin_set.shape[0], 1]),
                inds_notin_set, axis=1))
            qoi_combs = np.array_split(qoi_combs, comm.size)
        else:
            qoi_combs = None

        # Scatter them throughout the processors
        qoi_combs = comm.scatter(qoi_combs, root=0)

        # For each combination, compute the average measure(skewness) and add
        # the set to good_sets if it is less than measskew_tol
        for qoi_set in range(len(qoi_combs)):
            count_qois += 1
            curr_set = util.fix_dimensions_vector_2darray(qoi_combs[qoi_set])\
                .transpose()
            if measure is False:
                (current_measskew, singvals) = calculate_avg_skewness(input_set,
                                                                      qoi_combs[qoi_set])
            else:
                (current_measskew, singvals) = calculate_avg_measure(input_set,
                                                                     qoi_combs[qoi_set])

            # If its a good set, add it to good_sets
            if current_measskew < measskew_tol:
                good_sets = np.append(good_sets, curr_set, axis=0)

                # If the average skewness is less than the maxskewness
                # in our best_sets, add it to best_sets
                if current_measskew < best_sets[-1, 0]:
                    best_sets[-1, :] = np.append(np.array([current_measskew]),
                                                 qoi_combs[qoi_set])
                    order = best_sets[:, 0].argsort()
                    best_sets = best_sets[order]

                    # Store the corresponding singular values
                    optsingvals_tensor[:, :, -1] = singvals
                    optsingvals_tensor = optsingvals_tensor[:, :, order]

    # Wait for all processes to get to this point
    comm.Barrier()

    # Gather the best sets and skewness values from each processor
    good_sets = comm.gather(good_sets, root=0)
    best_sets = np.array(comm.gather(best_sets, root=0))
    count_qois = np.array(comm.gather(count_qois, root=0))

    # Find the num_optsets_return smallest skewness from all processors
    if comm.rank == 0:

        # Organize the best sets
        best_sets = best_sets.reshape(num_optsets_return *
                                      comm.size, num_qois_return + 1)
        [_, uniq_inds_best] = np.unique(best_sets[:, 0], return_index=True)
        best_sets = best_sets[uniq_inds_best, :]
        best_sets = best_sets[best_sets[:, 0].argsort()]
        best_sets = best_sets[:num_optsets_return, :]

        # Organize the good sets
        good_sets_new = np.zeros([1, num_qois_return])
        for each in good_sets:
            good_sets_new = np.append(good_sets_new, each[1:], axis=0)
        good_sets = good_sets_new

        logging.info('Possible sets of QoIs of size {} : {}'.format(
            good_sets.shape[1], np.sum(count_qois)))
        logging.info('Good sets of QoIs of size {} : {}'.format(
            good_sets.shape[1], good_sets.shape[0] - 1))

    comm.Barrier()
    best_sets = comm.bcast(best_sets, root=0)
    good_sets = comm.bcast(good_sets, root=0)

    return (good_sets[1:].astype(int), best_sets, optsingvals_tensor)


def chooseOptQoIs_large(input_set, qoiIndices=None, max_qois_return=None,
                        num_optsets_return=None, inner_prod_tol=None, measskew_tol=None,
                        measure=False, remove_zeros=True):
    r"""
    Given gradient vectors at some points (centers) in the input space, a large
    set of QoIs to choose from, and the number of desired QoIs to return, this
    method returns the set of optimal QoIs of size 2, 3, ...  max_qois_return
    to use in the inverse problem by choosing the sets with the smallest
    average measure(skewness).

    :param input_set: The input sample set.  Make sure the attribute _jacobians
        is not None.
    :type input_set: :class:`~bet.sample.sample_set`
    :param qoiIndices: Set of QoIs to consider from input_set._jacobians.
        Default is xrange(0, input_set._jacobians.shape[1])
    :type qoiIndices: :class:`np.ndarray` of size (1, num QoIs to consider)
    :param int max_qois_return: Maximum number of desired QoIs to use in the
        inverse problem.  Default is input_dim
    :param int num_optsets_return: Number of best sets to return
        Default is 10
    :param float inner_prod_tol: Maximum acceptable average inner product
        between two QoI maps.
    :param float measskew_tol: Throw out all sets of QoIs with average
        measure(skewness) number greater than this.
    :param boolean measure: If measure is True, use ``calculate_avg_measure``
        to determine optimal QoIs, else use ``calculate_avg_skewness``
    :param boolean remove_zeros: If True, ``find_unique_vecs`` will remove any
        QoIs that have a zero gradient vector at atleast one point in
        :math:`\Lambda`.

    :rtype: tuple
    :returns: (measure_skewness_indices_mat, optsingvals) where
        measure_skewness_indices_mat has shape (num_optsets_return,
        num_qois_return+1) and optsingvals has shape (num_centers,
        num_qois_return, num_optsets_return)

    """
    (best_sets, _) = chooseOptQoIs_large_verbose(input_set, qoiIndices,
                                                 max_qois_return, num_optsets_return, inner_prod_tol, measskew_tol, measure,
                                                 remove_zeros)

    return best_sets


def chooseOptQoIs_large_verbose(input_set, qoiIndices=None,
                                max_qois_return=None, num_optsets_return=None, inner_prod_tol=None,
                                measskew_tol=None, measure=False, remove_zeros=True):
    r"""
    Given gradient vectors at some points (centers) in the parameter space, a
    large set of QoIs to choose from, and the number of desired QoIs to return,
    this method return the set of optimal QoIs of size 1, 2, ... max_qois_return
    to use in the inverse problem by choosing the set with smallext average
    skewness.  Also a tensor that represents the singular values of the
    matrices formed by the gradient vectors of the optimal QoIs at each center
    is returned.

    :param input_set: The input sample set.  Make sure the attribute _jacobians
        is not None.
    :type input_set: :class:`~bet.sample.sample_set`
    :param qoiIndices: Set of QoIs to consider from G.  Default is
        xrange(0, G.shape[1]).
    :type qoiIndices: :class:`np.ndarray` of size (1, num QoIs to consider)
    :param int max_qois_return: Maximum number of desired QoIs to use in the
        inverse problem.  Default is input_dim.
    :param int num_optsets_return: Number of best sets to return.  Default is
        10.
    :param float inner_prod_tol: Throw out one vectors from each pair of
        QoIs that has average inner product greater than this.  Default is 0.9.
    :param float measskew_tol: Throw out all sets of QoIs with average
        measure(skewness) number greater than this.  Default is max_float.
    :param boolean measure: If measure is True, use ``calculate_avg_measure``
        to determine optimal QoIs, else use ``calculate_avg_skewness``
    :param boolean remove_zeros: If True, ``find_unique_vecs`` will remove any
        QoIs that have a zero gradient vector at atleast one point in
        :math:`\Lambda`.

    :rtype: tuple
    :returns: (measure_skewness_indices_mat, optsingvals) where
        measure_skewness_indices_mat has shape (num_optsets_return,
        num_qois_return+1) and optsingvals is a list where each element has
        shape (num_centers, num_qois_return, num_optsets_return).
        num_qois_return will change for each element of the list.

    """
    input_dim = input_set._dim
    if input_set._jacobians is None:
        raise ValueError("You must have jacobians to use this method.")
    if qoiIndices is None:
        qoiIndices = range(0, input_set._jacobians.shape[1])
    if max_qois_return is None:
        max_qois_return = input_dim
    if num_optsets_return is None:
        num_optsets_return = 10
    if inner_prod_tol is None:
        inner_prod_tol = 1.0
    if measskew_tol is None:
        measskew_tol = np.inf

    # Find the unique QoIs to consider
    unique_indices = find_unique_vecs(input_set, inner_prod_tol, qoiIndices,
                                      remove_zeros)
    if comm.rank == 0:
        logging.info('Unique Indices are : {}'.format(unique_indices))

    good_sets_curr = util.fix_dimensions_vector_2darray(unique_indices)
    best_sets = []
    optsingvals_list = []

    # Given good sets of QoIs of size (n - 1), find the good sets of size n
    for qois_return in range(2, max_qois_return + 1):
        (good_sets_curr, best_sets_curr, optsingvals_tensor_curr) = \
            find_good_sets(input_set, good_sets_curr, unique_indices,
                           num_optsets_return, measskew_tol, measure)
        best_sets.append(best_sets_curr)
        optsingvals_list.append(optsingvals_tensor_curr)
        if comm.rank == 0:
            logging.info(best_sets_curr)

    return (best_sets, optsingvals_list)
